from .node import UIGroup, UINode
import json
import rados

from gwcli.utils import console_message, os_cmd
import ceph_iscsi_config.settings as settings
from ceph_iscsi_config.utils import human_size


class CephGroup(UIGroup):
    """
    define an object to represent the ceph cluster. The methods use librados
    which means the host will need a valid ceph.conf and a valid keyring.
    """

    help_intro = '''
                 The ceph component of the shell is intended to provide
                 you with an overview of the ceph cluster. Information is
                 initially gathered when you start this shell, but can be
                 refreshed later using the 'refresh' subcommand. Data is shown
                 that covers the health of the ceph cluster(s), together with
                 an overview of the rados pools and overall topology.

                 The pools section is useful when performing allocation tasks
                 since it provides the current state of available space within
                 the pool(s), together with the current over-commit percentage.
                 '''

    def __init__(self, parent):
        UIGroup.__init__(self, 'cluster', parent)

        self.logger.debug("Adding ceph cluster '{}' to the UI"
                          .format(settings.config.cluster_name))
        self.cluster = CephCluster(self,
                                   settings.config.cluster_name,
                                   settings.config.cephconf,
                                   settings.config.cluster_client_name)

    @staticmethod
    def valid_conf(config_file):
        """
        Dummy check function
        :param config_file: (str) config file path
        :return: TRUE
        """

        return True

    def ui_command_refresh(self):
        """
        refresh command updates the health and capacity state of the ceph
        meta data shown within the interface
        """
        self.refresh()
        self.logger.info("ok")

    def refresh(self):

        for cluster in self.children:
            cluster.update_state()
            cluster.pools.refresh()

    def summary(self):
        """
        return the number of clusters
        :return:
        """
        return "Clusters: {}".format(len(self.children)), None


class CephCluster(UIGroup):

    def __init__(self, parent, cluster_name, conf_file, client_name):

        self.conf = conf_file
        self.client_name = client_name
        self.cluster_name = cluster_name
        UIGroup.__init__(self, cluster_name, parent)

        self.ceph_status = {}
        self.health_status = ''
        self.health_list = []
        self.version = self.cluster_version

        self.pools = CephPools(self)

        self.update_state()

        self.topology = CephTopology(self)

    def ui_command_refresh(self):
        """
        refresh command updates the health and capacity state of the ceph
        meta data shown within the interface
        """

        self.refresh()
        self.logger.info("ok")

    def ui_command_info(self):

        color = {"HEALTH_OK": "green",
                 "HEALTH_WARN": "yellow",
                 "HEALTH_ERR": "red"}

        status = self.ceph_status
        # backward compatibility (< octopus)
        if 'osdmap' in status['osdmap']:
            osdmap = status['osdmap']['osdmap']
        else:
            osdmap = status['osdmap']
        output = "Cluster name: {}\n".format(self.cluster_name)
        output += "Ceph version: {}\n".format(self.version)
        output += "Health      : {}\n".format(self.health_status)
        if self.health_status != 'HEALTH_OK':
            output += " - {}\n".format(','.join(self.health_list))

        # backward compatibility (< octopus)
        if 'mons' in status['monmap']:
            num_mons = len(status['monmap']['mons'])
        else:
            num_mons = status['monmap']['num_mons']
        num_quorum = len(status['quorum_names'])
        num_out_of_quorum = num_mons - num_quorum
        q_str = "quorum {}, out of quorum: {}".format(num_quorum, num_out_of_quorum)
        output += "\nMONs : {:>4} ({})\n".format(self.topology.num_mons,
                                                 q_str)
        output += "OSDs : {:>4} ({} up, {} in)\n".format(self.topology.num_osds,
                                                         osdmap['num_up_osds'],
                                                         osdmap['num_in_osds'])
        output += "Pools: {:>4}\n".format(self.num_pools)
        raw = status['pgmap']['bytes_total']
        output += "Raw capacity: {}\n".format(human_size(raw))
        output += "\nConfig : {}\n".format(self.conf)
        output += "Client Name: {}\n".format(self.client_name)

        console_message(output, color=color[self.health_status])

    @property
    def cluster_version(self):

        vers_out = os_cmd("ceph -c {} -n {} version".format(self.conf, self.client_name))

        # RHEL packages include additional info, that we don't need
        version_str = vers_out.split()[2].decode('utf-8')
        return '.'.join(version_str.split('.')[:3])

    def update_state(self):

        self.logger.debug("Querying ceph for state information")
        with rados.Rados(conffile=self.conf, name=self.client_name) as cluster:
            cluster.wait_for_latest_osdmap()

            cmd = {'prefix': 'status', 'format': 'json'}
            ret, buf_s, out = cluster.mon_command(json.dumps(cmd), b'')

        status = json.loads(buf_s)
        self.ceph_status = status

        # assume a luminous or above ceph cluster
        self.health_status = status['health'].get('status',
                                                  status['health'].get('overall_status'))
        self.health_list = [check.get('summary').get('message') for key, check in
                            status['health'].get('checks', {}).items()
                            if check.get('summary', {}).get('message')]

    @property
    def num_pools(self):
        return len([pool for pool in self.pools.children])

    def refresh(self):

        self.update_state()
        self.pools.refresh()

    def summary(self):

        color_lookup = {"HEALTH_OK": True,
                        "HEALTH_WARN": None,
                        "HEALTH_ERR": False}

        return self.health_status, color_lookup[self.health_status]

    @property
    def healthy_mon(self):
        """
        Return the first mon in quorum state
        :return: (str) name of the 1st monitor in a healthy state
        """

        quorum_mons = self.ceph_status.get('quorum_names', [])
        if quorum_mons:
            return quorum_mons[0]
        else:
            return "UNKNOWN"


class CephPools(UIGroup):
    help_intro = '''
                 Each pool within the ceph cluster is shown with the following
                 metrics;

                 - Commit .... this is a total of the logical space that has
                               been requested for all rbd images defined to
                               the gateways

                 - Avail ..... 'avail' shows the actual space that is available
                               for allocation after taking into account the
                               protection scheme of the pool (e.g. replication
                               level)

                 - Used ...... shows the physical space that has been consumed
                               within the pool

                 - Commit% ... is a ratio of the logical space allocated to
                               clients over the amount of space that can be
                               allocated. So when this value is <=100% the
                               physical backing store  capacity is available.
                               However, if this ratio is > 100%, you are
                               overcommiting capacity. Being able to overcommit
                               is a benefit of Ceph's thin provisioning - BUT
                               you must keep an eye on the capacity to protect
                               against out of space scenarios.

                 '''

    def __init__(self, parent):
        UIGroup.__init__(self, 'pools', parent)

        self.pool_lookup = {}  # pool_name -> pool object hash
        self.populate()

    def populate(self):

        # existing_pools = [pool.name for pool in self.children]

        # get a breakdown of the osd's to retrieve the pool types
        # i.e. is it replica 3 or EC 4+2 etc
        # This is overkill and hacky, but the bindings don't provide
        # pool type information...so it's SLEDGEHAMMER meets NUT time
        self.logger.debug("Fetching ceph osd information")
        with rados.Rados(conffile=self.parent.conf, name=self.parent.client_name) as cluster:
            cluster.wait_for_latest_osdmap()

            cmd = {'prefix': 'osd dump', 'format': 'json'}
            rc, buf_s, out = cluster.mon_command(json.dumps(cmd), b'')

        pools = {}
        for pool in json.loads(buf_s)['pools']:
            name = pool['pool_name']
            pools[name] = pool

        for pool_name in pools:
            # if pool_name not in existing_pools:
            new_pool = RadosPool(self, pool_name, pools[pool_name])
            self.pool_lookup[pool_name] = new_pool

    def refresh(self):

        self.logger.debug("Gathering pool stats for cluster "
                          "'{}'".format(self.parent.name))

        # unfortunately the rados python api does not expose all the needed
        # metrics through an ioctx call - specifically pool size is missing,
        # so stats need to be gathered at this level through the mon_command
        # interface, and pushed down to the child objects. Having a refresh
        # method within the child object would have been preferred!
        with rados.Rados(conffile=self.parent.conf, name=self.parent.client_name) as cluster:
            cluster.wait_for_latest_osdmap()

            cmd = {'prefix': 'df', 'format': 'json'}
            rc, buf_s, out = cluster.mon_command(json.dumps(cmd), b'')

            if rc == 0:
                pool_info = json.loads(buf_s)
                for pool_data in pool_info['pools']:
                    pool_name = pool_data['name']
                    self.pool_lookup[pool_name].update(pool_data)

    def summary(self):
        return "Pools: {}".format(self.parent.num_pools), True


class RadosPool(UINode):
    display_attributes = ["name", "commit", "overcommit_PCT", "max_bytes",
                          "used_bytes", "type", "desc"]

    def __init__(self, parent, pool_name, pool_md):
        UINode.__init__(self, pool_name, parent)
        self.pool_md = pool_md
        pool_type = {1: ("x{}".format(self.pool_md['size']),
                         'replicated'),
                     3: ("{}+{}".format(self.pool_md['min_size'],
                                        self.pool_md['size'] - self.pool_md['min_size']),
                         "erasure")}
        self.desc, self.type = pool_type[self.pool_md['type']]

    def _calc_overcommit(self):
        root = self.parent.parent.parent.parent
        potential_demand = 0
        for pool_child in root.disks.children:
            if pool_child.name == self.name:
                for image_child in pool_child.children:
                    potential_demand += image_child.size

        self.commit = potential_demand
        if self.max_bytes == 0:
            self.overcommit_PCT = 0
        else:
            self.overcommit_PCT = int(
                (potential_demand / float(self.max_bytes)) * 100)

    def update(self, pool_metadata):

        self.max_bytes = pool_metadata['stats']['max_avail']
        self.used_bytes = pool_metadata['stats']['bytes_used']

        self._calc_overcommit()

    def summary(self):
        msg = ["({})".format(self.desc)]
        msg.append("Commit: {}/{} ({}%)".format(human_size(self.commit),
                                                human_size(self.max_bytes),
                                                self.overcommit_PCT))
        msg.append("Used: {}".format(human_size(self.used_bytes)))

        return ', '.join(msg), True


class CephTopology(UINode):
    def __init__(self, parent):
        UINode.__init__(self, 'topology', parent)

        # backward compatibility (< octopus)
        if 'osdmap' in self.parent.ceph_status['osdmap']:
            self.num_osds = self.parent.ceph_status['osdmap']['osdmap']['num_osds']
        else:
            self.num_osds = self.parent.ceph_status['osdmap']['num_osds']

        # backward compatibility (< octopus)
        if 'mons' in self.parent.ceph_status['monmap']:
            self.num_mons = len(self.parent.ceph_status['monmap']['mons'])
        else:
            self.num_mons = self.parent.ceph_status['monmap']['num_mons']

    def summary(self):
        msg = ["OSDs: {}".format(self.num_osds)]
        msg.append("MONs: {}".format(self.num_mons))
        return ','.join(msg), True



import re

from gwcli.node import UIGroup, UINode
from gwcli.utils import response_message, APIRequest
import ceph_iscsi_config.settings as settings


class HostGroups(UIGroup):

    help_intro = '''
                 Hosts groups provide a more convenient way of managing multiple
                 hosts that require access to the same set of LUNs. The host
                 group 'policy' defines the clients and the LUNs (rbd images)
                 that should be associated together.

                 There are two commands used to manage the host group

                 create <group_name>
                 delete <group_name>

                 Since the same disks will be seen by multiple systems, you
                 should only use this feature for hosts that are cluster aware.
                 Failing to adhere to this constraint is likely to result in
                 **data loss**

                 Once a group has been created you can associate clients and
                 LUNs to the group with the 'host' and 'disk' sub-commands.

                 Note that a client can only belong to a single group definition,
                 but a disk can be defined across several groups.

    '''

    group_name_length = 32

    def __init__(self, parent):
        UIGroup.__init__(self, 'host-groups', parent)

        # record the shortcut
        shortcut = self.shell.prefs['bookmarks'].get('host-groups', None)
        if not shortcut or shortcut is not self.path:

            self.shell.prefs['bookmarks']['host-groups'] = self.path
            self.shell.prefs.save()

        self.load()

    def load(self):

        target_iqn = self.parent.name

        for child in self.children:
            self.delete(child)

        if target_iqn in self.get_ui_root().config['targets']:
            target_config = self.get_ui_root().config['targets'][target_iqn]
            groups = target_config['groups']
            for group_name in groups:
                HostGroup(self, group_name, groups[group_name])

    @property
    def groups(self):
        return [child.name for child in self.children]

    def summary(self):
        return "Groups : {}".format(len(self.children)), None

    def ui_command_create(self, group_name):
        """
        Create a host group definition. Group names can be use up to 32
        alphanumeric characters, including '_', '-' and '@'. Note that once a
        group is created it can not be renamed.
        """
        self.logger.debug("CMD: ../host-groups/ create {}".format(group_name))

        if group_name in self.groups:
            self.logger.error("Group {} already defined".format(group_name))
            return

        grp_regex = re.compile(
            r"^[\w\@\-\_]{{1,{}}}$".format(HostGroups.group_name_length))
        if not grp_regex.search(group_name):
            self.logger.error("Invalid group name - max of {} chars of "
                              "alphanumeric and -,_,@ "
                              "characters".format(HostGroups.group_name_length))
            return

        target_iqn = self.parent.name

        # this is a new group
        group_api = ('{}://{}:{}/api/hostgroup/'
                     '{}/{}'.format(self.http_mode,
                                    "localhost",
                                    settings.config.api_port,
                                    target_iqn,
                                    group_name))

        api = APIRequest(group_api)
        api.put()
        if api.response.status_code != 200:
            self.logger.error("Failed : "
                              "{}".format(response_message(api.response,
                                                           self.logger)))
            return

        self.logger.debug('Adding group to the UI')
        HostGroup(self, group_name)

        self.logger.info('ok')

        # Switch to the new group
        return self.ui_command_cd(group_name)

    def ui_command_delete(self, group_name):
        """
        Delete a host group definition. The delete process will remove the group
        definition and remove the group name association within any client.

        Note the deletion of a group will not remove the lun masking already
        defined to clients. If this is desired, it will need to be performed
        manually once the group is deleted.
        """

        self.logger.debug("CMD: ../host-groups/ delete {}".format(group_name))

        if group_name not in self.groups:
            self.logger.error("Group '{}' does not exist".format(group_name))
            return

        target_iqn = self.parent.name

        # OK, so the group exists...
        group_api = ('{}://{}:{}/api/hostgroup/'
                     '{}/{}'.format(self.http_mode,
                                    "localhost",
                                    settings.config.api_port,
                                    target_iqn,
                                    group_name))

        api = APIRequest(group_api)
        api.delete()
        if api.response.status_code != 200:
            self.logger.error("failed to delete group '{}'".format(group_name))
            return

        self.logger.debug("removing group from the UI")
        child = [child for child in self.children
                 if child.name == group_name][0]
        self.delete(child)

        self.logger.info('ok')

    def delete(self, child):

        target_iqn = self.parent.name

        client_group = child._get_client_group(target_iqn)
        client_map = client_group.client_map
        group_clients = [client_iqn for client_iqn in client_map
                         if client_map[client_iqn].group_name == child.name]

        for iqn in group_clients:

            self.logger.debug("removing group name from {}".format(iqn))
            client_map[iqn].group_name = ''

        self.remove_child(child)


class HostGroup(UIGroup):

    help_intro = '''
                 A host group provides a simple way to manage the LUN masking
                 of a number of iscsi clients as a single unit. The host group
                 contains hosts (iscsi clients) and disks (rbd images).

                 Once a host is defined to a group, it's lun masking must be
                 managed through the group. In fact attempts to manage the
                 disks of a client directly are blocked.

                 The following commands enable you to manage the membership of
                 the host group.

                 e.g.
                 host add|remove iqn.1994-05.com.redhat:rh7-client
                 disk add|remove rbd/disk_1
    '''

    valid_actions = ['add', 'remove']

    def __init__(self, parent, group_name, group_settings={}):
        UIGroup.__init__(self, group_name, parent)

        self.name = group_name

        for disk in group_settings.get('disks', []):
            HostGroupMember(self, 'disk', disk)
        for member in group_settings.get('members', []):
            HostGroupMember(self, 'host', member)

    def ui_command_host(self, action, client_iqn):
        """
        use the 'host' sub-command to add and remove hosts from a host group.
        Adding a host will automatically map the host group's disks to that
        specific host. Removing a host however, does not change the hosts
        disk masking - it simply removes the host from group.

        e.g.
        host add|remove iqn.1994-05.com.redhat:rh7-client
        """

        if action not in HostGroup.valid_actions:
            self.logger.error("Invalid request - must be "
                              "host add|remove <client_iqn>")
            return

        target_iqn = self.parent.parent.name

        # basic checks
        client_group = self._get_client_group(target_iqn)
        client_map = client_group.client_map
        if client_iqn not in client_map:
            self.logger.error("'{}' is not managed by a "
                              "group".format(client_iqn))
            return

        current_group = client_map[client_iqn].group_name
        if action == 'add' and current_group:
            self.logger.error("'{}' already belongs to "
                              "'{}'".format(client_iqn,
                                            current_group))
            return
        elif action == 'remove' and current_group != self.name:
            self.logger.error("'{}' does not belong to this "
                              "group".format(client_iqn))
            return

        # Basic checks passed, hand-off to the API now
        group_api = ('{}://{}:{}/api/hostgroup/'
                     '{}/{}'.format(self.http_mode,
                                    "localhost",
                                    settings.config.api_port,
                                    target_iqn,
                                    self.name))

        api_vars = {"action": action,
                    "members": client_iqn}

        api = APIRequest(group_api, data=api_vars)
        api.put()
        self.logger.debug("- api call responded "
                          "{}".format(api.response.status_code))
        if api.response.status_code != 200:
            self.logger.error("Failed :"
                              "{}".format(response_message(api.response,
                                                           self.logger)))
            return

        # group updated, so update the UI
        self.logger.debug("Updating the UI")
        if action == 'add':
            HostGroupMember(self, 'host', client_iqn)
            self.update_clients_UI([client_iqn], target_iqn)

        elif action == 'remove':
            child = [child for child in self.children
                     if child.name == client_iqn][0]
            self.delete(child)

        self.logger.info('ok')

    def delete(self, child):
        target_iqn = self.parent.parent.name

        if child.member_type == 'host':
            client_group = self._get_client_group(target_iqn)
            client = client_group.client_map[child.name]
            client.group_name = ''

        self.remove_child(child)

    def ui_command_disk(self, action, disk_name):
        """
        use the 'disk' sub-command to add or remove a disk from a specific
        host group. Removing disks should be done with care, as the remove
        operation will be executed across all hosts defined to the host group.

        e.g.
        disk add|remove rbd/disk_1
        """

        if action not in HostGroup.valid_actions:
            self.logger.error("Invalid request - must be "
                              "disk add|remove <disk_image>")
            return

        target_iqn = self.parent.parent.name

        # simple sanity checks
        # 1. does the disk exist in the configuration
        ui_root = self.get_ui_root()
        all_pools = ui_root.disks.children
        all_disks = []
        for current_pool in all_pools:
            for current_disk in current_pool.children:
                all_disks.append(current_disk)
        if disk_name not in [disk.image_id for disk in all_disks]:
            self.logger.error("Disk '{}' is not defined within the "
                              "configuration".format(disk_name))
            return

        # 2. For an 'add' request, the disk must not already be in the host
        # group. Whereas, for a remove request the disk must exist.
        if action == 'add':
            if disk_name in self.disks:
                self.logger.error("'{}' is already defined to this "
                                  "host-group".format(disk_name))
                return
        else:
            if disk_name not in self.disks:
                self.logger.error("'{}' is not a member of this "
                                  "group".format(disk_name))
                return

        mapped_disks = [mapped_disk.name
                        for mapped_disk in self.parent.parent.target_disks.children]
        if disk_name not in mapped_disks:
            rc = self.parent.parent.target_disks.add_disk(disk_name, None, None)
            if rc == 0:
                self.logger.debug("disk auto-map successful")
            else:
                self.logger.error("disk auto-map failed({}), try "
                                  "using the /iscsi-target/<iqn>/disks add "
                                  "command".format(rc))
                return

        # Basic checks passed, hand-off to the API
        group_api = ('{}://{}:{}/api/hostgroup/'
                     '{}/{}'.format(self.http_mode,
                                    "localhost",
                                    settings.config.api_port,
                                    target_iqn,
                                    self.name))

        api_vars = {"action": action,
                    "disks": disk_name}

        api = APIRequest(group_api, data=api_vars)
        api.put()
        self.logger.debug("- api call responded {}".format(api.response.status_code))
        if api.response.status_code != 200:
            self.logger.error("Failed: "
                              "{}".format(response_message(api.response,
                                                           self.logger)))
            return

        # group updated, so update the host-groups UI elements
        self.logger.debug("Updating the UI")
        if action == 'add':
            HostGroupMember(self, 'disk', disk_name)
        elif action == 'remove':
            child = [child for child in self.children
                     if child.name == disk_name][0]
            self.delete(child)

        self.update_clients_UI(self.members, target_iqn)

        self.logger.info('ok')

    @property
    def members(self):
        return [child.name for child in self.children
                if child.member_type == 'host']

    @property
    def disks(self):
        return [child.name for child in self.children
                if child.member_type == 'disk']

    def _get_client_group(self, target_iqn):
        ui_root = self.get_ui_root()
        iscsi_target = [t for t in list(ui_root.target.children)
                        if t.name == target_iqn][0]
        return iscsi_target.client_group

    def update_clients_UI(self, client_list, target_iqn):
        self.logger.debug("rereading the config object")
        root = self.get_ui_root()
        config = root._get_config()
        clients = config['targets'][target_iqn]['clients']

        client_group = self._get_client_group(target_iqn)     # Clients Object
        clients_to_update = [client for client in client_group.children
                             if client.name in client_list]

        # refresh the client with the new config
        self.logger.debug("resync'ing client lun maps")
        client_map = client_group.client_map
        for client in clients_to_update:
            client_map[client.client_iqn].group_name = self.name
            client.drop_luns()
            client.luns = clients[client.client_iqn].get('luns', {})
            client.refresh_luns()

    def summary(self):
        counts = {'disk': 0, 'host': 0}
        for child in self.children:
            counts[child.member_type] += 1
        return "Hosts: {}, Disks: {}".format(counts['host'],
                                             counts['disk']), \
               None


class HostGroupMember(UINode):
    help_intro = '''
                    The entries here show the hosts and disks that are held
                    within a specific host group definition. Care should be
                    taken when removing disks from a host group, as the remove
                    operation will be performed across each client within the
                    group.
                 '''

    def __init__(self, parent, member_type, name):
        UINode.__init__(self, name, parent)
        self.member_type = member_type

    def summary(self):
        return "{}".format(self.member_type), True



import rados
import socket
import time
import json
import traceback

from ceph_iscsi_config.backstore import USER_RBD
import ceph_iscsi_config.settings as settings
from ceph_iscsi_config.utils import encryption_available, get_time


class ConfigTransaction(object):

    def __init__(self, cfg_type, element_name, txn_action='add', initial_value=None):

        self.type = cfg_type
        self.action = txn_action
        self.item_name = element_name

        init_state = {} if initial_value is None else initial_value
        self.item_content = init_state

    def __repr__(self):
        return str(self.__dict__)


class CephCluster(object):

    def __init__(self):

        self.error = False
        self.error_msg = ''
        self.cluster = None

        conf = settings.config.cephconf

        try:
            self.cluster = rados.Rados(conffile=conf,
                                       name=settings.config.cluster_client_name)
        except rados.Error as err:
            self.error = True
            self.error_msg = "Invaid cluster_client_name or setting in {} - {}".format(conf, err)
            return

        try:
            self.cluster.connect()
        except rados.Error as err:
            self.error = True
            self.error_msg = "Unable to connect to the cluster (keyring missing?) - {}".format(err)

    def __del__(self):
        if self.cluster:
            self.cluster.shutdown()

    def shutdown(self):
        self.cluster.shutdown()


class Config(object):

    seed_config = {"disks": {},
                   "gateways": {},
                   "targets": {},
                   "discovery_auth": {'username': '',
                                      'password': '',
                                      'password_encryption_enabled': False,
                                      'mutual_username': '',
                                      'mutual_password': '',
                                      'mutual_password_encryption_enabled': False},
                   "version": 11,
                   "epoch": 0,
                   "created": '',
                   "updated": ''
                   }

    lock_time_limit = 30

    def __init__(self, logger, cfg_name='gateway.conf', pool=None):
        self.logger = logger
        self.config_name = cfg_name
        if pool is None:
            pool = settings.config.pool
        self.pool = pool
        self.ceph = None
        self.error = False
        self.reset = False
        self.error_msg = ""
        self.txn_list = []
        self.config_locked = False

        self.ceph = CephCluster()
        if self.ceph.error:
            self.error = True
            self.error_msg = self.ceph.error_msg
            return

        if self.init_config():
            self.config = self.get_config()
            self._upgrade_config()
            self.changed = False

    def _read_config_object(self, ioctx):
        """
        Return config string from the config object. The string is checked to
        see if it's valid json. If it's not the read is likely to be a against
        the object while it's being updated by another host - if this happens,
        we wait and reread until we get valid json.
        :param ioctx: rados ioctx
        :return: (str) current string.
        """

        try:
            size, mtime = ioctx.stat(self.config_name)
        except rados.ObjectNotFound:
            self.logger.error("_read_config_object object not found")
            raise
        else:
            self.logger.debug("_read_config_object reading the config object")
            size += 1
            cfg_str = ioctx.read(self.config_name, length=size)
            if cfg_str:
                valid = False
                while not valid:
                    try:
                        json.loads(cfg_str)
                    except ValueError:
                        #
                        self.logger.debug("_read_config_object not valid json, rereading")
                        time.sleep(1)
                        size, mtime = ioctx.stat(self.config_name)
                        cfg_str = ioctx.read(self.config_name, length=size)
                    else:
                        valid = True

        return cfg_str

    def _open_ioctx(self):
        try:
            self.logger.debug("(_open_ioctx) Opening connection to {} pool".format(self.pool))
            ioctx = self.ceph.cluster.open_ioctx(self.pool)
        except rados.ObjectNotFound:
            self.error = True
            self.error_msg = "'{}' pool does not exist!".format(self.pool)
            self.logger.error("(_open_ioctx) {} does not exist".format(self.pool))
            raise
        self.logger.debug("(_open_ioctx) connection opened")
        return ioctx

    def _get_ceph_config(self):

        cfg_dict = {}

        ioctx = self._open_ioctx()
        cfg_data = self._read_config_object(ioctx)
        ioctx.close()

        if not cfg_data:
            # attempt to read the object got nothing which means it's empty
            # so we seed the object
            self.logger.debug("(_get_rbd_config) config object is empty..seeding it")
            self._seed_rbd_config()
            if self.error:
                self.logger.error("(Config._get_rbd_config) Unable to seed the config object")
                return {}
            else:
                cfg_data = json.dumps(Config.seed_config)

        self.logger.debug("(_get_rbd_config) config object contains '{}'".format(cfg_data))

        cfg_dict = json.loads(cfg_data)

        return cfg_dict

    def needs_hostname_update(self):
        if self.config['version'] == 9:
            # No gateway has been updated yet.
            return True

        updated = self.config.get('gateways_upgraded')
        if updated is None:
            # Everything has been updated or we are < 9
            return False

        if socket.getfqdn() in updated:
            return False

        return True

    def _upgrade_config(self):
        update_hostname = self.needs_hostname_update()

        if self.config['version'] >= Config.seed_config['version'] and not update_hostname:
            return

        if self.config['version'] <= 2:
            self.add_item("groups", element_name=None, initial_value={})
            self.update_item("version", element_name=None, element_value=3)

        if self.config['version'] == 3:
            iqn = self.config['gateways'].get('iqn', None)
            gateways = {}
            portals = {}

            self.add_item("targets", None, {})
            self.add_item('discovery_auth', None, {
                'chap': '',
                'chap_mutual': ''
            })

            if iqn:
                for host, gateway_v3 in self.config['gateways'].items():
                    if isinstance(gateway_v3, dict):
                        portal = gateway_v3
                        portal.pop('iqn')
                        active_luns = portal.pop('active_luns')
                        updated = portal.pop('updated', None)
                        created = portal.pop('created', None)
                        gateway = {
                            'active_luns': active_luns
                        }
                        if created:
                            gateway['created'] = created
                        if updated:
                            gateway['updated'] = updated
                        gateways[host] = gateway
                        portals[host] = portal
                for _, client in self.config['clients'].items():
                    client.pop('created', None)
                    client.pop('updated', None)
                    client['auth']['chap_mutual'] = ''
                for _, group in self.config['groups'].items():
                    group.pop('created', None)
                    group.pop('updated', None)
                target = {
                    'disks': list(self.config['disks'].keys()),
                    'clients': self.config['clients'],
                    'portals': portals,
                    'groups': self.config['groups'],
                    'controls': self.config.get('controls', {}),
                    'ip_list': self.config['gateways']['ip_list']
                }
                self.add_item("targets", iqn, target)
                self.update_item("targets", iqn, target)

            self.update_item("gateways", None, gateways)

            if 'controls' in self.config:
                self.del_item('controls', None)
            self.del_item('clients', None)
            self.del_item('groups', None)

            self.update_item("version", None, 4)

        if self.config['version'] == 4:
            for disk_id, disk in self.config['disks'].items():
                disk['backstore'] = USER_RBD
                self.update_item("disks", disk_id, disk)
            self.update_item("version", None, 5)

        if self.config['version'] == 5:
            for target_iqn, target in self.config['targets'].items():
                target['acl_enabled'] = True
                self.update_item("targets", target_iqn, target)
            self.update_item("version", None, 6)

        if self.config['version'] == 6:
            new_disks = {}
            old_disks = []
            for disk_id, disk in self.config['disks'].items():
                disk['backstore_object_name'] = disk_id
                new_disk_id = disk_id.replace('.', '/')
                new_disks[new_disk_id] = disk
                old_disks.append(disk_id)
            for old_disk_id in old_disks:
                self.del_item('disks', old_disk_id)
            for new_disk_id, new_disk in new_disks.items():
                self.add_item("disks", new_disk_id, new_disk)
            for iqn, target in self.config['targets'].items():
                new_disk_ids = []
                for disk_id in target['disks']:
                    new_disk_id = disk_id.replace('.', '/')
                    new_disk_ids.append(new_disk_id)
                target['disks'] = new_disk_ids
                for _, client in target['clients'].items():
                    new_luns = {}
                    for lun_id, lun in client['luns'].items():
                        new_lun_id = lun_id.replace('.', '/')
                        new_luns[new_lun_id] = lun
                    client['luns'] = new_luns
                for _, group in target['groups'].items():
                    new_group_disks = {}
                    for group_disk_id, group_disk in group['disks'].items():
                        new_group_disk_id = group_disk_id.replace('.', '/')
                        new_group_disks[new_group_disk_id] = group_disk
                        group['disks'] = new_group_disks
                self.update_item("targets", iqn, target)
            self.update_item("version", None, 7)

        if self.config['version'] == 7:
            if '/' in self.config['discovery_auth']['chap']:
                duser, dpassword = self.config['discovery_auth']['chap'].split('/', 1)
            else:
                duser = ''
                dpassword = ''
            self.config['discovery_auth']['username'] = duser
            self.config['discovery_auth']['password'] = dpassword
            self.config['discovery_auth']['password_encryption_enabled'] = False
            self.config['discovery_auth'].pop('chap', None)
            if '/' in self.config['discovery_auth']['chap_mutual']:
                dmuser, dmpassword = self.config['discovery_auth']['chap_mutual'].split('/', 1)
            else:
                dmuser = ''
                dmpassword = ''
            self.config['discovery_auth']['mutual_username'] = dmuser
            self.config['discovery_auth']['mutual_password'] = dmpassword
            self.config['discovery_auth']['mutual_password_encryption_enabled'] = False
            self.config['discovery_auth'].pop('chap_mutual', None)
            self.update_item("discovery_auth", None, self.config['discovery_auth'])
            for target_iqn, target in self.config['targets'].items():
                for _, client in target['clients'].items():
                    if '/' in client['auth']['chap']:
                        user, password = client['auth']['chap'].split('/', 1)
                    else:
                        user = ''
                        password = ''
                    client['auth']['username'] = user
                    client['auth']['password'] = password
                    client['auth']['password_encryption_enabled'] = \
                        (len(password) > 16 and encryption_available())
                    client['auth'].pop('chap', None)
                    if '/' in client['auth']['chap_mutual']:
                        muser, mpassword = client['auth']['chap_mutual'].split('/', 1)
                    else:
                        muser = ''
                        mpassword = ''
                    client['auth']['mutual_username'] = muser
                    client['auth']['mutual_password'] = mpassword
                    client['auth']['mutual_password_encryption_enabled'] = \
                        (len(mpassword) > 16 and encryption_available())
                    client['auth'].pop('chap_mutual', None)

                self.update_item("targets", target_iqn, target)
            self.update_item("version", None, 8)

        if self.config['version'] == 8:
            for target_iqn, target in self.config['targets'].items():
                for _, portal in target['portals'].items():
                    portal['portal_ip_addresses'] = [portal['portal_ip_address']]
                    portal.pop('portal_ip_address')
                self.update_item("targets", target_iqn, target)
            self.update_item("version", None, 9)

        if self.config['version'] == 9 or update_hostname:
            # temporary field to store the gateways already upgraded from v9 to v10
            gateways_upgraded = self.config.get('gateways_upgraded')
            if not gateways_upgraded:
                gateways_upgraded = []
                self.add_item('gateways_upgraded', None, gateways_upgraded)
            this_shortname = socket.gethostname().split('.')[0]
            this_fqdn = socket.getfqdn()
            if this_fqdn not in gateways_upgraded:
                gateways_config = self.config['gateways']
                gateway_config = gateways_config.get(this_shortname)
                if gateway_config:
                    gateways_config.pop(this_shortname)
                    gateways_config[this_fqdn] = gateway_config
                    self.update_item("gateways", None, gateways_config)
                for target_iqn, target in self.config['targets'].items():
                    portals_config = target['portals']
                    portal_config = portals_config.get(this_shortname)
                    if portal_config:
                        portals_config.pop(this_shortname)
                        portals_config[this_fqdn] = portal_config
                        self.update_item("targets", target_iqn, target)
                for disk_id, disk in self.config['disks'].items():
                    if disk.get('allocating_host') == this_shortname:
                        disk['allocating_host'] = this_fqdn
                    if disk.get('owner') == this_shortname:
                        disk['owner'] = this_fqdn
                    self.update_item("disks", disk_id, disk)
                gateways_upgraded.append(this_fqdn)
                self.update_item("gateways_upgraded", None, gateways_upgraded)

            if any(gateway_name not in gateways_upgraded
                   for gateway_name in self.config['gateways'].keys()):
                self.logger.debug("gateways upgraded to 10: {}".
                                  format(gateways_upgraded))
            else:
                self.del_item("gateways_upgraded", None)

            if self.config['version'] == 9:
                # Upgrade from v9 to v10 is still in progress. Update the
                # version now, so we can update the other config fields and
                # setup the target to execute IO while the other gws upgrade.
                self.update_item("version", None, 10)

        # Currently, the versions below do not rely on fields being updated
        # in the 9->10 upgrade which needs to execute on every node before
        # completing. If this changes, we will need to fix how we handle
        # rolling upgrades, so new versions have access to the updated fields
        # on all gws before completing the upgrade.
        if self.config['version'] == 10:
            for target_iqn, target in self.config['targets'].items():
                target['auth'] = {
                    'username': '',
                    'password': '',
                    'password_encryption_enabled': False,
                    'mutual_username': '',
                    'mutual_password': '',
                    'mutual_password_encryption_enabled': False
                }
                disks = {}
                for disk_index, disk in enumerate(sorted(target['disks'])):
                    disks[disk] = {
                        'lun_id': disk_index
                    }
                target['disks'] = disks
                self.update_item("targets", target_iqn, target)
            self.update_item("version", None, 11)

        self.commit("retain")

    def init_config(self):
        try:
            ioctx = self._open_ioctx()
        except rados.ObjectNotFound:
            return False

        try:
            with rados.WriteOpCtx(ioctx) as op:
                # try to exclusively create the config object
                op.new(rados.LIBRADOS_CREATE_EXCLUSIVE)
                ioctx.operate_write_op(op, self.config_name)
                self.logger.debug("(init_config) created empty config object")
        except rados.ObjectExists:
            self.logger.debug("(init_config) using pre existing config object")
        ioctx.close()
        return True

    def get_config(self):
        return self._get_ceph_config()

    def lock(self):

        ioctx = self._open_ioctx()

        secs = 0
        self.logger.debug("config.lock attempting to acquire lock on {}".format(self.config_name))
        while secs < Config.lock_time_limit:
            try:
                ioctx.lock_exclusive(self.config_name, 'lock', 'config')
                self.config_locked = True
                break
            except (rados.ObjectBusy, rados.ObjectExists):
                self.logger.debug("(Config.lock) waiting for excl lock on "
                                  "{} object".format(self.config_name))
                time.sleep(1)
                secs += 1

        if secs >= Config.lock_time_limit:
            self.error = True
            self.error_msg = ("Timed out ({}s) waiting for excl "
                              "lock on {} object".format(Config.lock_time_limit, self.config_name))
            self.logger.error("(Config.lock) {}".format(self.error_msg))

        ioctx.close()

    def unlock(self):
        ioctx = self._open_ioctx()

        self.logger.debug("config.unlock releasing lock on {}".format(self.config_name))
        try:
            ioctx.unlock(self.config_name, 'lock', 'config')
            self.config_locked = False
        except Exception:
            self.error = True
            self.error_msg = ("Unable to unlock {} - {}".format(self.config_name,
                                                                traceback.format_exc()))
            self.logger.error("(Config.unlock) {}".format(self.error_msg))

        ioctx.close()

    def _seed_rbd_config(self):

        ioctx = self._open_ioctx()

        self.lock()
        if self.error:
            return

        # if the config object is empty, seed it - if not just leave as is
        cfg_data = self._read_config_object(ioctx)
        if not cfg_data:
            self.logger.debug("_seed_rbd_config found empty config object")
            seed_now = Config.seed_config
            seed_now['created'] = get_time()
            seed = json.dumps(seed_now, sort_keys=True, indent=4, separators=(',', ': '))
            ioctx.write_full(self.config_name, seed.encode('utf-8'))
            ioctx.set_xattr(self.config_name, "epoch", "0".encode('utf-8'))
            self.changed = True

        self.unlock()

    def refresh(self):
        self.logger.debug("config refresh - current config is {}".format(self.config))
        self.config = self.get_config()
        self._upgrade_config()

    def add_item(self, cfg_type, element_name=None, initial_value=None):
        now = get_time()

        if element_name:
            # ensure the initial state for this item has a 'created' date/time value
            if isinstance(initial_value, dict):
                if 'created' not in initial_value:
                    initial_value['created'] = now

            if initial_value is None:
                init_state = {"created": now}
            else:
                init_state = initial_value

            self.config[cfg_type][element_name] = init_state

            if isinstance(init_state, str) and 'created' not in self.config[cfg_type]:
                self.config[cfg_type]['created'] = now
                # add a separate transaction to capture the creation date to the section
                txn = ConfigTransaction(cfg_type, 'created', initial_value=now)
                self.txn_list.append(txn)

        else:
            # new section being added to the config object
            self.config[cfg_type] = initial_value
            init_state = initial_value
            txn = ConfigTransaction(cfg_type, None, initial_value=initial_value)
            self.txn_list.append(txn)

        self.logger.debug("(Config.add_item) config updated to {}".format(self.config))
        self.changed = True

        txn = ConfigTransaction(cfg_type, element_name, initial_value=init_state)
        self.txn_list.append(txn)

    def del_item(self, cfg_type, element_name):
        self.changed = True
        if element_name:
            del self.config[cfg_type][element_name]
        else:
            del self.config[cfg_type]
        self.logger.debug("(Config.del_item) config updated to {}".format(self.config))

        txn = ConfigTransaction(cfg_type, element_name, 'delete')
        self.txn_list.append(txn)

    def update_item(self, cfg_type, element_name, element_value):
        now = get_time()

        if element_name:
            current_values = self.config[cfg_type][element_name]
            self.logger.debug("prior to update, item contains {}".format(current_values))
            if isinstance(element_value, dict):
                merged = current_values.copy()
                new_dict = element_value
                new_dict['updated'] = now
                merged.update(new_dict)
                element_value = merged.copy()

            self.config[cfg_type][element_name] = element_value
        else:
            # update to a root level config element, like version
            self.config[cfg_type] = element_value

        self.logger.debug("(Config.update_item) config is {}".format(self.config))
        self.changed = True
        self.logger.debug("update_item: type={}, item={}, update={}".format(
            cfg_type, element_name, element_value))

        txn = ConfigTransaction(cfg_type, element_name, 'add')
        txn.item_content = element_value
        self.txn_list.append(txn)

    def set_item(self, cfg_type, element_name, element_value):
        self.logger.debug("(Config.update_item) config is {}".format(self.config))
        self.changed = True
        self.logger.debug("update_item: type={}, item={}, update={}".format(
            cfg_type, element_name, element_value))

        txn = ConfigTransaction(cfg_type, element_name, 'add')
        txn.item_content = element_value
        self.txn_list.append(txn)

    def _commit_rbd(self, post_action):

        ioctx = self._open_ioctx()

        if not self.config_locked:
            self.lock()
            if self.error:
                return

        # reread the config to account for updates made by other systems
        # then apply this hosts update(s)
        current_config = json.loads(self._read_config_object(ioctx))
        for txn in self.txn_list:

            self.logger.debug("_commit_rbd transaction shows {}".format(txn))
            if txn.action == 'add':         # add's and updates
                if txn.item_name:
                    current_config[txn.type][txn.item_name] = txn.item_content
                else:
                    current_config[txn.type] = txn.item_content

            elif txn.action == 'delete':
                if txn.item_name:
                    del current_config[txn.type][txn.item_name]
                else:
                    del current_config[txn.type]
            else:
                self.error = True
                self.error_msg = "Unknown transaction type ({}} encountered in " \
                                 "_commit_rbd".format(txn.action)

        if not self.error:
            if self.reset:
                current_config["epoch"] = 0
            else:
                # Python will switch from plain to long int automagically
                current_config["epoch"] += 1

            now = get_time()
            current_config['updated'] = now
            config_str = json.dumps(current_config)
            self.logger.debug("_commit_rbd updating config to {}".format(config_str))
            config_str_fmtd = json.dumps(current_config, sort_keys=True,
                                         indent=4, separators=(',', ': '))
            ioctx.write_full(self.config_name, config_str_fmtd.encode('utf-8'))
            ioctx.set_xattr(self.config_name, "epoch",
                            str(current_config["epoch"]).encode('utf-8'))
            del self.txn_list[:]                # empty the list of transactions

        self.unlock()
        ioctx.close()

        if post_action == 'close':
            self.ceph.shutdown()

    def commit(self, post_action='close'):
        self._commit_rbd(post_action)


def main():
    pass


if __name__ == '__main__':
    main()



from rtslib_fb.fabric import ISCSIFabricModule

from ceph_iscsi_config.client import CHAP
from ceph_iscsi_config.utils import encryption_available


class Discovery(object):

    @staticmethod
    def set_discovery_auth_lio(username, password, password_encryption_enabled, mutual_username,
                               mutual_password, mutual_password_encryption_enabled):
        iscsi_fabric = ISCSIFabricModule()
        if username == '':
            iscsi_fabric.clear_discovery_auth_settings()
        else:
            chap = CHAP(username, password, password_encryption_enabled)
            chap_mutual = CHAP(mutual_username, mutual_password,
                               mutual_password_encryption_enabled)
            iscsi_fabric.discovery_userid = chap.user
            iscsi_fabric.discovery_password = chap.password
            iscsi_fabric.discovery_mutual_userid = chap_mutual.user
            iscsi_fabric.discovery_mutual_password = chap_mutual.password
            iscsi_fabric.discovery_enable_auth = True

    @staticmethod
    def set_discovery_auth_config(username, password, mutual_username, mutual_password, config):
        encryption_enabled = encryption_available()
        discovery_auth_config = {
            'username': '',
            'password': '',
            'password_encryption_enabled': encryption_enabled,
            'mutual_username': '',
            'mutual_password': '',
            'mutual_password_encryption_enabled': encryption_enabled
        }
        if username != '':
            chap = CHAP(username, password, encryption_enabled)
            chap_mutual = CHAP(mutual_username, mutual_password, encryption_enabled)
            discovery_auth_config['username'] = chap.user
            discovery_auth_config['password'] = chap.encrypted_password(encryption_enabled)
            discovery_auth_config['mutual_username'] = chap_mutual.user
            discovery_auth_config['mutual_password'] = \
                chap_mutual.encrypted_password(encryption_enabled)
        config.update_item('discovery_auth', '', discovery_auth_config)



import threading
import time
import os

import rtslib_fb.tcm as tcm
from rtslib_fb.root import RTSRoot
from rtslib_fb.utils import fread

from .utils import this_host


class Metric(object):
    """ Metric object used to hold the metric, labels and value """

    def __init__(self, vhelp, vtype):
        self.var_help = vhelp
        self.var_type = vtype
        self.data = []

    def add(self, labels, value):
        _d = dict(labels=labels,
                  value=value)
        self.data.append(_d)


class TPGMapper(threading.Thread):
    """ thread which builds a list of LUNs mapped to a given TPG"""
    def __init__(self, tpg):
        self.tpg = tpg
        self.tpg_id = tpg.tag
        self.portal_ip = next(tpg.network_portals).ip_address
        self.owned_luns = dict()
        threading.Thread.__init__(self)

    def run(self):
        for lun in self.tpg.luns:
            if lun.alua_tg_pt_gp_name == 'ao':
                lun_name = lun.storage_object.name
                self.owned_luns[lun_name] = self.portal_ip


class GatewayStats(object):
    """ Gather and format gateway related performance data """

    def __init__(self):
        self.metrics = {}
        self._root = RTSRoot()

        # use utils.this_host
        self.gw_name = this_host()

    def formatted(self):
        s = ''
        for m_name in sorted(self.metrics.keys()):
            metric = self.metrics[m_name]
            s += "#HELP: {} {}\n".format(m_name,
                                         metric.var_help)
            s += "#TYPE: {} {}\n".format(m_name,
                                         metric.var_type)

            for v in metric.data:
                labels = []
                for n in v['labels'].items():
                    label_name = '{}='.format(n[0])
                    label_value = '"{}"'.format(n[1])

                    labels.append('{}{}'.format(label_name,
                                                label_value))

                s += "{}{{{}}} {}\n".format(m_name,
                                            ','.join(labels),
                                            v["value"])

        return s.rstrip()

    def collect(self):

        # the tcm module uses a global called bs_cache and performs lookups
        # against this to verify a storage object exists. However, if a change
        # is made the local copy of bs_cache in the rbd-target-gw scope is not
        # changed, so we reset it here to ensure it always starts empty
        tcm.bs_cache = {}

        stime = time.time()
        self._get_tpg()
        self._get_mapping()
        self._get_lun_sizes()
        self._get_lun_stats()
        self._get_client_details()
        etime = time.time()

        summary = Metric("time taken to scrape iscsi stats (secs)",
                         "gauge")
        labels = {"gw_name": self.gw_name}
        summary.add(labels, etime - stime)
        self.metrics['ceph_iscsi_scrape_duration_seconds'] = summary

    def _get_tpg(self):
        stat = Metric("target portal groups defined within gateway group",
                      "gauge")
        labels = {"gw_iqn": next(self._root.targets).wwn}
        v = len([tpg for tpg in self._root.tpgs])
        stat.add(labels, v)

        self.metrics["ceph_iscsi_gateway_tpg_total"] = stat

    def _get_mapping(self):
        mapping = Metric("LUN mapping state 0=unmapped, 1=mapped",
                         "gauge")
        mapped_devices = [l.tpg_lun.storage_object.name
                          for l in self._root.mapped_luns]

        tpg_mappers = []
        for tpg in self._root.tpgs:
            mapper = TPGMapper(tpg)
            mapper.start()
            tpg_mappers.append(mapper)

        for mapper in tpg_mappers:
            mapper.join()

        # merge the tpg lun maps
        all_devs = tpg_mappers[0].owned_luns.copy()
        for mapper in tpg_mappers[1:]:
            all_devs.update(mapper.owned_luns)

        for so in self._root.storage_objects:

            so_state = 1 if so.name in mapped_devices else 0
            owner = all_devs[so.name]
            mapping.add({"lun_name": so.name,
                         "gw_name": self.gw_name,
                         "gw_owner": owner}, so_state)

        self.metrics["ceph_iscsi_lun_mapped"] = mapping

    def _get_lun_sizes(self):
        size_bytes = Metric("LUN size (bytes)",
                            "gauge")
        for so in self._root.storage_objects:
            labels = {"lun_name": so.name,
                      "gw_name": self.gw_name}
            lun_size = so.size
            size_bytes.add(labels, lun_size)
        self.metrics["ceph_iscsi_lun_size_bytes"] = size_bytes

    def _get_lun_stats(self):
        iops = Metric("IOPS per LUN per client",
                      "counter")
        read_bytes = Metric("read bytes per LUN per client",
                            "counter")
        write_bytes = Metric("write bytes per LUN client",
                             "counter")

        for node_acl in self._root.node_acls:
            for lun in node_acl.mapped_luns:
                lun_path = lun.path
                lun_name = lun.tpg_lun.storage_object.name
                perf_labels = {"gw_name": self.gw_name,
                               "client_iqn": node_acl.node_wwn,
                               "lun_name": lun_name}

                lun_iops = int(fread(
                    os.path.join(lun_path,
                                 "statistics/scsi_auth_intr/num_cmds")))
                mbytes_read = int(fread(
                    os.path.join(lun_path,
                                 "statistics/scsi_auth_intr/read_mbytes")))
                mbytes_write = int(fread(
                    os.path.join(lun_path,
                                 "statistics/scsi_auth_intr/write_mbytes")))

                iops.add(perf_labels,
                         lun_iops)
                read_bytes.add(perf_labels,
                               mbytes_read * (1024 ** 2))
                write_bytes.add(perf_labels,
                                mbytes_write * (1024 ** 2))

        self.metrics["ceph_iscsi_lun_iops"] = iops
        self.metrics["ceph_iscsi_lun_read_bytes"] = read_bytes
        self.metrics["ceph_iscsi_lun_write_bytes"] = write_bytes

    def _get_client_details(self):
        logins = Metric("iscsi client session active (0=No, 1=Yes)",
                        "gauge")
        lun_map = Metric("LUN ID by client",
                         "gauge")
        logged_in_clients = [client['parent_nodeacl'].node_wwn
                             for client in self._root.sessions
                             if client['state'] == 'LOGGED_IN']

        for client in self._root.node_acls:

            login_labels = {"gw_name": self.gw_name,
                            "client_iqn": client.node_wwn
                            }

            v = 1 if client.node_wwn in logged_in_clients else 0
            logins.add(login_labels, v)

            for lun in client.mapped_luns:
                lun_labels = {"gw_name": self.gw_name,
                              "client_iqn": client.node_wwn,
                              "lun_name": lun.tpg_lun.storage_object.name}
                v = lun.mapped_lun
                lun_map.add(lun_labels, v)

        self.metrics["ceph_iscsi_client_login"] = logins
        self.metrics["ceph_iscsi_client_lun"] = lun_map



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import unittest

from ceph_iscsi_config.settings import Settings
from ceph_iscsi_config.target import GWTarget


class SettingsTest(unittest.TestCase):

    @staticmethod
    def _normalize(controls):
        return Settings.normalize_controls(controls, GWTarget.SETTINGS)

    def test_normalize_controls_int(self):
        self.assertEqual(
            SettingsTest._normalize({'dataout_timeout': 3}), {'dataout_timeout': 3})
        self.assertEqual(
            SettingsTest._normalize({'dataout_timeout': '3'}), {'dataout_timeout': 3})

        with self.assertRaises(ValueError) as cm:
            SettingsTest._normalize({'dataout_timeout': 1})
        self.assertEqual('expected integer >= 2 for dataout_timeout', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            SettingsTest._normalize({'dataout_timeout': 64})
        self.assertEqual('expected integer <= 60 for dataout_timeout', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            SettingsTest._normalize({'dataout_timeout': '64'})
        self.assertEqual('expected integer <= 60 for dataout_timeout', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            SettingsTest._normalize({'dataout_timeout': 'abc'})
        self.assertEqual('expected integer for dataout_timeout', str(cm.exception))

    def test_normalize_controls_yes_no(self):
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': 'Yes'}), {'immediate_data': True})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': 'yes'}), {'immediate_data': True})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': True}), {'immediate_data': True})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': 'True'}), {'immediate_data': True})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': 'true'}), {'immediate_data': True})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': '1'}), {'immediate_data': True})

        self.assertEqual(
            SettingsTest._normalize({'immediate_data': 'No'}), {'immediate_data': False})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': 'no'}), {'immediate_data': False})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': False}), {'immediate_data': False})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': 'False'}), {'immediate_data': False})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': 'false'}), {'immediate_data': False})
        self.assertEqual(
            SettingsTest._normalize({'immediate_data': '0'}), {'immediate_data': False})

        with self.assertRaises(ValueError) as cm:
            SettingsTest._normalize({'immediate_data': 'abc'})
        self.assertEqual('expected yes or no for immediate_data', str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            SettingsTest._normalize({'immediate_data': 123})
        self.assertEqual('expected yes or no for immediate_data', str(cm.exception))



import logging


def convert_str_to_bool(value):
    """
    Convert true/false/yes/no/1/0 to boolean
    """

    if isinstance(value, bool):
        return value

    value = str(value).lower()
    if value in ['1', 'true', 'yes']:
        return True
    elif value in ['0', 'false', 'no']:
        return False
    raise ValueError(value)


class Setting(object):
    def __init__(self, name, type_str, def_val):
        self.name = name
        self.type_str = type_str
        self.def_val = def_val

    def __contains__(self, key):
        return key == self.def_val


class BoolSetting(Setting):
    def __init__(self, name, def_val):
        super(BoolSetting, self).__init__(name, "bool", def_val)

    def to_str(self, norm_val):
        if norm_val:
            return "true"
        else:
            return "false"

    def normalize(self, raw_val):
        try:
            # for compat we also support Yes/No and 1/0
            return convert_str_to_bool(raw_val)
        except ValueError:
            raise ValueError("expected true or false for {}".format(self.name))


class LIOBoolSetting(BoolSetting):
    def __init__(self, name, def_val):
        super(LIOBoolSetting, self).__init__(name, def_val)

    def to_str(self, norm_val):
        if norm_val:
            return "yes"
        else:
            return "no"

    def normalize(self, raw_val):
        try:
            # for compat we also support True/False and 1/0
            return convert_str_to_bool(raw_val)
        except ValueError:
            raise ValueError("expected yes or no for {}".format(self.name))


class ListSetting(Setting):
    def __init__(self, name, def_val):
        super(ListSetting, self).__init__(name, "list", def_val)

    def to_str(self, norm_val):
        return str(norm_val)

    def normalize(self, raw_val):
        return raw_val.split(',') if raw_val else []


class StrSetting(Setting):
    def __init__(self, name, def_val):
        super(StrSetting, self).__init__(name, "str", def_val)

    def to_str(self, norm_val):
        return str(norm_val)

    def normalize(self, raw_val):
        return str(raw_val)


class IntSetting(Setting):
    def __init__(self, name, min_val, max_val, def_val):
        self.min_val = min_val
        self.max_val = max_val
        super(IntSetting, self).__init__(name, "int", def_val)

    def to_str(self, norm_val):
        return str(norm_val)

    def normalize(self, raw_val):
        try:
            val = int(raw_val)
        except ValueError:
            raise ValueError("expected integer for {}".format(self.name))

        if val < self.min_val:
            raise ValueError("expected integer >= {} for {}".
                             format(self.min_val, self.name))
        if val > self.max_val:
            raise ValueError("expected integer <= {} for {}".
                             format(self.max_val, self.name))
        return val


class EnumSetting(Setting):
    def __init__(self, name, valid_vals, def_val):
        if len(valid_vals) == 0:
            raise ValueError("Invalid enum. There must be at least one valid value.")

        valid_type = type(valid_vals[0])
        if valid_type is not int and valid_type is not str:
            raise ValueError("Invalid enum. Items must be str or int. Got {}".
                             format(valid_type))

        for i in valid_vals:
            if valid_type != type(i):
                raise ValueError("Invalid enum. All items must be the same type."
                                 "Found {} and {}".format(type(i), valid_type))

        self.valid_vals = valid_vals
        super(EnumSetting, self).__init__(name, "enum", def_val)

    def to_str(self, norm_val):
        return str(norm_val)

    def normalize(self, raw_val):
        if isinstance(self.valid_vals[0], str):
            val = str(raw_val)
        else:
            val = int(raw_val)

        if val not in self.valid_vals:
            raise ValueError("expected {} for {} found {}".
                             format(self.valid_vals, self.name, raw_val))

        return val


CLIENT_SETTINGS = {
    "dataout_timeout": IntSetting("dataout_timeout", 2, 60, 20),
    "nopin_response_timeout": IntSetting("nopin_response_timeout", 3, 60, 5),
    "nopin_timeout": IntSetting("nopin_timeout", 3, 60, 5),
    "cmdsn_depth": IntSetting("cmdsn_depth", 1, 512, 128)}

TGT_SETTINGS = {
    # client settings you can also set at the ceph-iscsi target level
    "dataout_timeout": IntSetting("dataout_timeout", 2, 60, 20),
    "nopin_response_timeout": IntSetting("nopin_response_timeout", 3, 60, 5),
    "nopin_timeout": IntSetting("nopin_timeout", 3, 60, 5),
    "cmdsn_depth": IntSetting("cmdsn_depth", 1, 512, 128),
    # lio tpg settings
    "immediate_data": LIOBoolSetting("immediate_data", True),
    "initial_r2t": LIOBoolSetting("initial_r2t", True),
    "max_outstanding_r2t": IntSetting("max_outstanding_r2t", 1, 65535, 1),
    "first_burst_length": IntSetting("first_burst_length", 512, 16777215, 262144),
    "max_burst_length": IntSetting("max_burst_length", 512, 16777215, 524288),
    "max_recv_data_segment_length": IntSetting("max_recv_data_segment_length",
                                               512, 16777215, 262144),
    "max_xmit_data_segment_length": IntSetting("max_xmit_data_segment_length",
                                               512, 16777215, 262144)}

SYS_SETTINGS = {
    "cluster_name": StrSetting("cluster_name", "ceph"),
    "pool": StrSetting("pool", "rbd"),
    "cluster_client_name": StrSetting("cluster_client_name", "client.admin"),
    "time_out": IntSetting("time_out", 1, 600, 30),
    "api_host": StrSetting("api_host", "::"),
    "api_port": IntSetting("api_port", 1, 65535, 5000),
    "api_secure": BoolSetting("api_secure", True),
    "api_ssl_verify": BoolSetting("api_ssl_verify", False),
    "loop_delay": IntSetting("loop_delay", 1, 60, 2),
    "trusted_ip_list": ListSetting("trusted_ip_list", []),  # comma separate list of IPs
    "api_user": StrSetting("api_user", "admin"),
    "api_password": StrSetting("api_password", "admin"),
    "ceph_user": StrSetting("ceph_user", "admin"),
    "debug": BoolSetting("debug", False),
    "minimum_gateways": IntSetting("minimum_gateways", 1, 9999, 2),
    "ceph_config_dir": StrSetting("ceph_config_dir", '/etc/ceph'),
    "priv_key": StrSetting("priv_key", 'iscsi-gateway.key'),
    "pub_key": StrSetting("pub_key", 'iscsi-gateway-pub.key'),
    "prometheus_exporter": BoolSetting("prometheus_exporter", True),
    "prometheus_port": IntSetting("prometheus_port", 1, 65535, 9287),
    "prometheus_host": StrSetting("prometheus_host", "::"),
    "logger_level": IntSetting("logger_level", logging.DEBUG, logging.CRITICAL,
                               logging.DEBUG),
    # TODO: This is under sys for compat. It is not settable per device/backend
    # type yet.
    "alua_failover_type": EnumSetting("alua_failover_type",
                                      ["implicit", "explicit"], "implicit")}

TCMU_SETTINGS = {
    "max_data_area_mb": IntSetting("max_data_area_mb", 1, 2048, 8),
    "qfull_timeout": IntSetting("qfull_timeout", 0, 600, 5),
    "osd_op_timeout": IntSetting("osd_op_timeout", 0, 600, 30),
    "hw_max_sectors": IntSetting("hw_max_sectors", 1, 8192, 1024)}



#!/usr/bin/python

import logging
import logging.handlers
from logging.handlers import RotatingFileHandler

from flask import Flask, Response

from ceph_iscsi_config.metrics import GatewayStats

import ceph_iscsi_config.settings as settings

# Create a flask instance
app = Flask(__name__)


@app.route("/", methods=["GET"])
def prom_root():
    """ handle the '/' endpoint - just redirect point the user at /metrics"""
    return '''<!DOCTYPE html>
    <html>
        <head><title>Ceph/iSCSI Prometheus Exporter</title></head>
        <body>
            <h1>Ceph/iSCSI Prometheus Exporter</h1>
            <p><a href='/metrics'>Metrics</a></p>
        </body>
    </html>'''


@app.route("/metrics", methods=["GET"])
def prom_metrics():
    """ Collect the stats and send back to the caller"""

    stats = GatewayStats()
    stats.collect()

    return Response(stats.formatted(),
                    content_type="text/plain")


def main():

    if settings.config.prometheus_exporter:

        logger.info("Integrated Prometheus exporter is enabled")
        # starting a flask instance will occupy the main thread

        # Attach the werkzeug log to the handlers defined in the outer scope
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.DEBUG)
        log.addHandler(file_handler)
        log.addHandler(syslog_handler)

        app.run(host=settings.config.prometheus_host,
                port=settings.config.prometheus_port,
                debug=False,
                threaded=True)

    else:
        logger.info("Integrated Prometheus exporter is disabled")


if __name__ == '__main__':

    settings.init()
    logger_level = logging.getLevelName(settings.config.logger_level)

    # setup syslog handler to help diagnostics
    logger = logging.getLogger('rbd-target-gw')
    logger.setLevel(logging.DEBUG)

    # syslog (systemctl/journalctl messages)
    syslog_handler = logging.handlers.SysLogHandler(address='/dev/log')
    syslog_handler.setLevel(logging.INFO)
    syslog_format = logging.Formatter("%(message)s")
    syslog_handler.setFormatter(syslog_format)

    # file target - more verbose logging for diagnostics
    file_handler = RotatingFileHandler('/var/log/rbd-target-gw/rbd-target-gw.log',
                                       maxBytes=5242880,
                                       backupCount=7)
    file_handler.setLevel(logger_level)
    file_format = logging.Formatter("%(asctime)s [%(levelname)8s] - %(message)s")
    file_handler.setFormatter(file_format)

    logger.addHandler(syslog_handler)
    logger.addHandler(file_handler)

    main()



from rtslib_fb.alua import ALUATargetPortGroup

import ceph_iscsi_config.settings as settings
from ceph_iscsi_config.utils import CephiSCSIInval


def alua_format_group_name(tpg, failover_type, is_owner):
    if is_owner:
        return "ao"

    if failover_type == "explicit":
        return "standby{}".format(tpg.tag)
    else:
        return "ano{}".format(tpg.tag)


def alua_create_ao_group(so, tpg, group_name):
    alua_tpg = ALUATargetPortGroup(so, group_name, tpg.tag)
    alua_tpg.alua_support_active_optimized = 1
    alua_tpg.alua_access_state = 0

    return alua_tpg


def alua_create_implicit_group(tpg, so, group_name, is_owner):
    if is_owner:
        alua_tpg = alua_create_ao_group(so, tpg, group_name)
    else:
        alua_tpg = ALUATargetPortGroup(so, group_name, tpg.tag)
        alua_tpg.alua_access_state = 1

    alua_tpg.alua_support_active_nonoptimized = 1
    alua_tpg.alua_access_type = 1
    # Just make sure we get to at least attempt one op for the failover
    # process.
    alua_tpg.implicit_trans_secs = settings.config.osd_op_timeout + 15
    return alua_tpg


def alua_create_explicit_group(tpg, so, group_name, is_owner):
    if is_owner:
        alua_tpg = alua_create_ao_group(so, tpg, group_name)
        alua_tpg.preferred = 1
    else:
        alua_tpg = ALUATargetPortGroup(so, group_name, tpg.tag)

    alua_tpg.alua_support_standby = 1
    # Use Explicit but also set the Implicit bit so we can
    # update the kernel from configfs.
    alua_tpg.alua_access_type = 3
    # start ports in Standby, and let the initiator drive the initial
    # transition to AO.
    alua_tpg.alua_access_state = 2
    return alua_tpg


def alua_create_group(failover_type, tpg, so, is_owner):
    group_name = alua_format_group_name(tpg, failover_type, is_owner)

    if failover_type == "explicit":
        alua_tpg = alua_create_explicit_group(tpg, so, group_name, is_owner)
    elif failover_type == "implicit":
        # tmp drop down to implicit. Next patch will check for "implicit"
        # and add error handling up the stack if the failover_type is invalid.
        alua_tpg = alua_create_implicit_group(tpg, so, group_name, is_owner)
    else:
        raise CephiSCSIInval("Invalid failover type {}".format(failover_type))

    alua_tpg.alua_support_active_optimized = 1
    alua_tpg.alua_support_offline = 0
    alua_tpg.alua_support_unavailable = 0
    alua_tpg.alua_support_transitioning = 1
    alua_tpg.nonop_delay_msecs = 0

    return alua_tpg



import json
import threading

from gwcli.node import UIGroup, UINode, UIRoot
from gwcli.hostgroup import HostGroups
from gwcli.storage import Disks, TargetDisks
from gwcli.client import Clients
from gwcli.utils import (response_message, GatewayAPIError,
                         GatewayError, APIRequest, console_message, get_config,
                         refresh_control_values)

import ceph_iscsi_config.settings as settings
from ceph_iscsi_config.utils import (normalize_ip_address, this_host)
from ceph_iscsi_config.target import GWTarget
from ceph_iscsi_config.client import CHAP

from gwcli.ceph import CephGroup

from rtslib_fb.utils import normalize_wwn, RTSLibError

# FIXME - code is using a self signed cert common across all gateways
# the embedded urllib3 package will issue warnings when ssl cert validation is
# disabled - so this disable_warnings stops the user interface from being
# bombed
from requests.packages import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class ISCSIRoot(UIRoot):

    def __init__(self, shell, scan_threads=1, endpoint=None):
        UIRoot.__init__(self, shell)

        self.error = False
        self.error_msg = ''
        self.interactive = True           # default interactive mode
        self.scan_threads = scan_threads

        if settings.config.api_secure:
            self.http_mode = 'https'
        else:
            self.http_mode = 'http'

        if endpoint is None:

            self.local_api = ('{}://localhost:{}/'
                              'api'.format(self.http_mode,
                                           settings.config.api_port))

        else:
            self.local_api = endpoint

        self.config = {}
        # Establish the root nodes within the UI, for the different components

        self.disks = Disks(self)
        self.ceph = CephGroup(self)
        self.target = ISCSITargets(self)

    def refresh(self):
        self.config = self._get_config()

        if not self.error:

            if 'disks' in self.config:
                self.disks.refresh(self.config['disks'])
            else:
                self.disks.refresh({})

            if 'gateways' in self.config:
                self.target.gateway_group = self.config['gateways']
            else:
                self.target.gateway_group = {}

            self.target.refresh(self.config['targets'], self.config['discovery_auth'])

            self.ceph.refresh()

        else:
            # Unable to get the config, tell the user and exit the cli
            self.logger.critical("Unable to access the configuration "
                                 "object : {}".format(self.error_msg))
            raise GatewayError

    def _get_config(self, endpoint=None):

        if not endpoint:
            endpoint = self.local_api

        api = APIRequest(endpoint + "/config")
        api.get()

        if api.response.status_code == 200:
            try:
                return api.response.json()
            except Exception:
                self.error = True
                self.error_msg = "Malformed REST API response"
                return {}
        else:
            self.error = True
            self.error_msg = "REST API failure, code : " \
                             "{}".format(api.response.status_code)
            return {}

    def export_copy(self, config):

        fmtd_config = json.dumps(config, sort_keys=True,
                                 indent=4, separators=(',', ': '))
        print(fmtd_config)

    def ui_command_export(self, mode='copy'):
        """
        Print the configuration in a format that can be used as a backup.

        The export command supports two modes:

        copy - This prints the internal configuration. It can used for backup
               or for support requests.
        """

        valid_modes = ['copy']

        self.logger.debug("CMD: export mode={}".format(mode))

        if mode not in valid_modes:
            self.logger.error("Invalid export mode requested - supported "
                              "modes are: {}".format(','.join(valid_modes)))
            return

        current_config = self._get_config()

        if mode == 'copy':
            self.export_copy(current_config)

    def ui_command_info(self):
        self.logger.debug("CMD: info")

        if settings.config.trusted_ip_list:
            display_ips = ','.join(settings.config.trusted_ip_list)
        else:
            display_ips = 'None'

        console_message("HTTP mode          : {}".format(self.http_mode))
        console_message("Rest API port      : {}".format(settings.config.api_port))
        console_message("Local endpoint     : {}".format(self.local_api))
        console_message("Local Ceph Cluster : {}".format(settings.config.cluster_name))
        console_message("2ndary API IP's    : {}".format(display_ips))


class ISCSITargets(UIGroup):
    help_intro = '''
                 The iscsi-target group defines the ISCSI Target that the
                 group of gateways will be known as by iSCSI initiators (clients).

                 Only one iSCSI target is allowed, but each target can consist
                 of 2-4 gateway nodes. Multiple gateways are needed to deliver
                 high availability storage to the iSCSI client.

                 '''

    def __init__(self, parent):
        UIGroup.__init__(self, 'iscsi-targets', parent)
        self.gateway_group = {}
        self.auth = None

    def ui_command_create(self, target_iqn):
        """
        Create an iSCSI target. This target is defined across all gateway nodes,
        providing the client with a single 'image' for iscsi discovery.
        """

        self.logger.debug("CMD: /iscsi create {}".format(target_iqn))

        # is the IQN usable?
        try:
            target_iqn, iqn_type = normalize_wwn(['iqn'], target_iqn)
        except RTSLibError:
            self.logger.error("IQN name '{}' is not valid for "
                              "iSCSI".format(target_iqn))
            return

        # 'safe' to continue with the definition
        self.logger.debug("Create an iscsi target definition in the UI")

        local_api = ('{}://localhost:{}/api/'
                     'target/{}'.format(self.http_mode,
                                        settings.config.api_port,
                                        target_iqn))

        api = APIRequest(local_api)
        api.put()

        if api.response.status_code == 200:
            self.logger.info('ok')
            # create the target entry in the UI tree
            target_exists = len([target for target in self.children
                                 if target.name == target_iqn]) > 0
            if not target_exists:
                Target(target_iqn, self)
        else:
            self.logger.error("Failed to create the target on the local node")

            raise GatewayAPIError("iSCSI target creation failed - "
                                  "{}".format(response_message(api.response,
                                                               self.logger)))

    def ui_command_delete(self, target_iqn):
        """
        Delete an iSCSI target.
        """

        self.logger.debug("CMD: /iscsi delete {}".format(target_iqn))

        try:
            target_iqn, iqn_type = normalize_wwn(['iqn'], target_iqn)
        except RTSLibError:
            self.logger.error("IQN name '{}' is not valid for "
                              "iSCSI".format(target_iqn))
            return

        gw_api = ('{}://localhost:{}/api/'
                  'target/{}'.format(self.http_mode,
                                     settings.config.api_port,
                                     target_iqn))

        api = APIRequest(gw_api)
        api.delete()

        if api.response.status_code == 200:
            self.logger.info('ok')
            # delete the target entry from the UI tree
            target_object = [target for target in self.children
                             if target.name == target_iqn][0]
            self.remove_child(target_object)
        else:
            self.logger.error("Failed - {}".format(response_message(api.response, self.logger)))

    def ui_command_clearconfig(self, confirm=None):
        """
        The 'clearconfig' command allows you to return the configuration to an
        unused state: LIO on each gateway will be cleared, and gateway
        definitions in the configuration object will be removed.

        > clearconfig confirm=true

        In order to run the clearconfig command, all clients and disks *must*
        have already have been removed.
        """

        self.logger.debug("CMD: clearconfig confirm={}".format(confirm))

        confirm = self.ui_eval_param(confirm, 'bool', False)
        if not confirm:
            self.logger.error("To clear the configuration you must specify "
                              "confirm=true")
            return

        # get a new copy of the config dict over the local API
        # check that there aren't any disks or client listed
        current_config = get_config()
        for target_iqn, target in current_config['targets'].items():
            num_clients = len(target['clients'].keys())
            if num_clients > 0:
                self.logger.error("{} - Clients({}) must be removed first"
                                  " before clearing the gateway "
                                  "configuration".format(target_iqn,
                                                         num_clients))
                return

        num_disks = len(current_config['disks'].keys())
        if num_disks > 0:
            self.logger.error("Disks({}) must be removed first"
                              " before clearing the gateway "
                              "configuration".format(num_disks))
            return

        for target_iqn, target in current_config['targets'].items():
            target_config = current_config['targets'][target_iqn]

            self.clear_config(target_config['portals'], target_iqn)

    def clear_config(self, gw_list, target_iqn):

        for gw_name in gw_list:

            gw_api = ('{}://{}:{}/api/'
                      '_target/{}'.format(self.http_mode,
                                          gw_name,
                                          settings.config.api_port,
                                          target_iqn))

            api = APIRequest(gw_api)
            api.delete()
            if api.response.status_code != 200:
                msg = response_message(api.response, self.logger)
                self.logger.error("Delete of {} failed : {}".format(gw_name,
                                                                    msg))
                raise GatewayAPIError
            else:
                self.logger.debug("- deleted {}".format(gw_name))

        # gateways removed, so lets delete the objects from the UI tree
        self.reset()

        # remove any bookmarks stored in the prefs.bin file
        if 'bookmarks' in self.shell.prefs:
            del self.shell.prefs['bookmarks']

        self.logger.info('ok')

    def ui_command_discovery_auth(self, username=None, password=None, mutual_username=None,
                                  mutual_password=None):
        """
        Discovery authentication can be set to use CHAP/CHAP_MUTUAL by supplying
        username, password, mutual_username, mutual_password

        Specifying 'nochap' will remove discovery authentication.

        e.g.
        auth username=<user> password=<pass> mutual_username=<m_user> mutual_password=<m_pass>

        """

        self.logger.warn("discovery username={}, password={}, mutual_username={}, "
                         "mutual_password={}".format(username, password, mutual_username,
                                                     mutual_password))

        self.logger.debug("CMD: /iscsi discovery_auth")

        if not username:
            self.logger.error("To set or reset discovery authentication, specify either "
                              "username=<user> password=<password> [mutual_username]=<user> "
                              "[mutual_password]=<password> or nochap")
            return

        if username == 'nochap':
            username = ''
            password = ''
            mutual_username = ''
            mutual_password = ''

        self.logger.debug("discovery auth to be set to username='{}', password='{}', "
                          "mutual_username='{}', mutual_password='{}'".format(username, password,
                                                                              mutual_username,
                                                                              mutual_password))

        api_vars = {
            "username": username,
            "password": password,
            "mutual_username": mutual_username,
            "mutual_password": mutual_password
        }
        discoveryauth_api = ('{}://localhost:{}/api/'
                             'discoveryauth'.format(self.http_mode,
                                                    settings.config.api_port))
        api = APIRequest(discoveryauth_api, data=api_vars)
        api.put()

        if api.response.status_code == 200:
            self._set_auth(username, password, mutual_username, mutual_password)
            self.logger.info('ok')
        else:
            self.logger.error("Error: {}".format(response_message(api.response, self.logger)))
            return

    def _set_auth(self, username, password, mutual_username, mutual_password):
        if mutual_username != '' and mutual_password != '':
            self.auth = "CHAP_MUTUAL"
        elif username != '' and password != '':
            self.auth = "CHAP"
        else:
            self.auth = None

    def refresh(self, targets, discovery_auth):

        self.logger.debug("Refreshing gateway & client information")
        self.reset()
        self._set_auth(discovery_auth['username'],
                       discovery_auth['password'],
                       discovery_auth['mutual_username'],
                       discovery_auth['mutual_password'])
        for target_iqn, target in targets.items():
            tgt = Target(target_iqn, self)
            tgt.controls = target['controls']

            tgt.gateway_group.load(target['portals'])
            tgt.target_disks.load(target['disks'])
            tgt.client_group.load(target['clients'])

    def summary(self):
        return "DiscoveryAuth: {}, Targets: {}".format(self.auth, len(self.children)), None


class Target(UINode):

    display_attributes = ["target_iqn", "auth", "control_values"]

    help_intro = '''
                 The iscsi target is the name that the group of gateways are
                 known as by the iscsi initiators (clients).
                 '''

    def __init__(self, target_iqn, parent):
        UIGroup.__init__(self, target_iqn, parent)
        self.target_iqn = target_iqn

        self.control_values = []
        self.controls = {}

        self.gateway_group = GatewayGroup(self)
        self.client_group = Clients(self)
        self.host_groups = HostGroups(self)
        self.target_disks = TargetDisks(self)

        config = self.parent.parent._get_config()
        if not config:
            self.logger.error("Unable to refresh local config")

        self.auth = config['targets'][target_iqn]['auth']
        # decode the chap password if necessary
        if 'username' in self.auth and 'password' in self.auth:
            self.chap = CHAP(self.auth['username'],
                             self.auth['password'],
                             self.auth['password_encryption_enabled'])
            self.auth['username'] = self.chap.user
            self.auth['password'] = self.chap.password
        else:
            self.auth['username'] = ''
            self.auth['password'] = ''

        # decode the chap_mutual password if necessary
        if 'mutual_username' in self.auth and 'mutual_password' in self.auth:
            self.chap_mutual = CHAP(self.auth['mutual_username'],
                                    self.auth['mutual_password'],
                                    self.auth['mutual_password_encryption_enabled'])
            self.auth['mutual_username'] = self.chap_mutual.user
            self.auth['mutual_password'] = self.chap_mutual.password
        else:
            self.auth['mutual_username'] = ''
            self.auth['mutual_password'] = ''

    def _get_controls(self):
        return self._controls.copy()

    def _set_controls(self, controls):
        self._controls = controls.copy()
        self.control_values = {}
        refresh_control_values(self.control_values, self.controls,
                               GWTarget.SETTINGS)

    controls = property(_get_controls, _set_controls)

    def summary(self):
        msg = []

        auth_text = "Auth: None"
        status = None
        if self.auth.get('mutual_username'):
            auth_text = "Auth: CHAP_MUTUAL"
            status = True
        elif self.auth.get('username'):
            auth_text = "Auth: CHAP"
            status = True
        msg.append(auth_text)

        msg.append("Gateways: {}".format(len(self.gateway_group.children)))

        return ", ".join(msg), status

    def ui_command_reconfigure(self, attribute, value):
        """
        The reconfigure command allows you to tune various gateway attributes.
        An empty value for an attribute resets the lun attribute to its
        default.
        attribute : attribute to reconfigure. supported attributes:
          - cmdsn_depth : integer 1 - 512
          - dataout_timeout : integer 2 - 60
          - nopin_response_timeout : integer 3 - 60
          - nopin_timeout : integer 3 - 60
          - immediate_data : [Yes|No]
          - initial_r2t : [Yes|No]
          - first_burst_length : integer 512 - 16777215
          - max_burst_length : integer 512 - 16777215
          - max_outstanding_r2t : integer 1 - 65535
          - max_recv_data_segment_length : integer 512 - 16777215
          - max_xmit_data_segment_length : integer 512 - 16777215
        value     : value of the attribute to reconfigure
        e.g.
        set cmdsn_depth
          - reconfigure attribute=cmdsn_depth value=128
        reset cmdsn_depth
          - reconfigure attribute=cmdsn_depth value=
        """
        if not GWTarget.SETTINGS.get(attribute):
            self.logger.error("supported attributes: {}".format(",".join(
                sorted(GWTarget.SETTINGS.keys()))))
            return

        # Issue the api request for the reconfigure
        gateways_api = ('{}://localhost:{}/api/'
                        'target/{}'.format(self.http_mode,
                                           settings.config.api_port,
                                           self.target_iqn))

        controls = {attribute: value}
        controls_json = json.dumps(controls)
        api_vars = {'mode': 'reconfigure', 'controls': controls_json}

        self.logger.debug("Issuing reconfigure request: controls={}".format(controls_json))
        api = APIRequest(gateways_api, data=api_vars)
        api.put()

        if api.response.status_code != 200:
            self.logger.error("Failed to reconfigure : "
                              "{}".format(response_message(api.response,
                                                           self.logger)))
            return

        config = self.parent.parent._get_config()
        if not config:
            self.logger.error("Unable to refresh local config")
        self.controls = config['targets'][self.target_iqn]['controls']

        self.logger.info('ok')

    def ui_command_auth(self, username=None, password=None, mutual_username=None,
                        mutual_password=None):
        """
        Target authentication can be set to use CHAP/CHAP_MUTUAL by supplying
        username, password, mutual_username, mutual_password

        e.g.
        auth username=<user> password=<pass> mutual_username=<m_user> mutual_password=<m_pass>

        username / mutual_username ... the username is 8-64 character string. Each character
                                       may either be an alphanumeric or use one of the following
                                       special characters .,:,-,@.
                                       Consider using the hosts 'shortname' or the initiators IQN
                                       value as the username

        password / mutual_password ... the password must be between 12-16 chars in length
                                       containing alphanumeric characters, plus the following
                                       special characters @,_,-,/
        """

        self.logger.debug("CMD: /iscsi-targets/<target_iqn> auth *")

        if not username:
            self.logger.error("To set authentication, specify "
                              "username=<user> password=<password> "
                              "[mutual_username]=<user> [mutual_password]=<password> "
                              "or nochap")
            return

        if username == 'nochap':
            username = ''
            password = ''
            mutual_username = ''
            mutual_password = ''

        self.logger.debug("auth to be set to username='{}', password='{}', mutual_username='{}', "
                          "mutual_password='{}'".format(username, password,
                                                        mutual_username, mutual_password))
        target_iqn = self.name

        api_vars = {
            "username": username,
            "password": password,
            "mutual_username": mutual_username,
            "mutual_password": mutual_password
        }

        targetauth_api = ('{}://localhost:{}/api/'
                          'targetauth/{}'.format(self.http_mode,
                                                 settings.config.api_port,
                                                 target_iqn))
        api = APIRequest(targetauth_api, data=api_vars)
        api.put()

        if api.response.status_code == 200:
            self.logger.debug("- target credentials updated")
            self.auth['username'] = username
            self.auth['password'] = password
            self.auth['mutual_username'] = mutual_username
            self.auth['mutual_password'] = mutual_password
            self.logger.info('ok')

        else:
            self.logger.error("Failed to update target auth: "
                              "{}".format(response_message(api.response,
                                                           self.logger)))
            return


class GatewayGroup(UIGroup):

    help_intro = '''
                 The gateway-group shows you the high level details of the
                 iscsi gateway nodes that have been configured. It also allows
                 you to add further gateways to the configuration, but this
                 requires the API service instance to be started on the new
                 gateway host

                 If in doubt, use Ansible :)
                 '''

    def __init__(self, parent):

        UIGroup.__init__(self, 'gateways', parent)

        self.thread_lock = threading.Lock()
        self.check_interval = 10           # check gateway state every 'n' secs
        self.last_state = 0

        # record the shortcut
        shortcut = self.shell.prefs['bookmarks'].get('gateways', None)
        if not shortcut or shortcut is not self.path:

            self.shell.prefs['bookmarks']['gateways'] = self.path
            self.shell.prefs.save()
            self.shell.log.debug("Bookmarked %s as %s."
                                 % (self.path, 'gateways'))

    @property
    def gateways_down(self):
        return len([gw for gw in self.children
                    if gw.state != 'UP'])

    def load(self, portal_group):
        for gateway_name in portal_group:
            Gateway(self, gateway_name, portal_group[gateway_name])

        self.check_gateways()

    def ui_command_info(self):
        '''
        List configured gateways.
        '''
        self.logger.debug("CMD: ../gateways/ info")

        for child in self.children:
            console_message(child)

    def check_gateways(self):

        check_thread = threading.Timer(self.check_interval,
                                       self.check_gateways)
        check_thread.daemon = True
        check_thread.start()
        self.refresh()

    def refresh(self, mode='auto'):

        self.thread_lock.acquire()

        if len(self.children) > 0:
            for gw in self.children:
                gw.refresh(mode)
        else:
            pass

        gateways_down = self.gateways_down
        if gateways_down != self.last_state:
            if gateways_down == 0:
                self.logger.info("\nAll gateways accessible")
            else:
                err_str = "gateway is" if gateways_down == 1 else "gateways are"
                self.logger.warning("\n{} {} inaccessible - updates will "
                                    "be disabled".format(gateways_down,
                                                         err_str))
            self.last_state = gateways_down

        self.thread_lock.release()

    def ui_command_refresh(self):
        """
        refresh allows you to refresh the connection status of each of the
        configured gateways (i.e. check the up/down state).
        """
        num_gw = len(self.children)
        if num_gw > 0:
            self.logger.debug("{} gateways to refresh".format(num_gw))
            self.refresh(mode='interactive')
        else:
            self.logger.error("No gateways to refresh")

    def ui_command_delete(self, gateway_name, confirm=None):
        """
        Delete a gateway from the group. This will stop and delete the target
        running on the gateway.

        If this is the last gateway the target is mapped to all objects added
        to it will be removed, and confirm=True is required.
        """

        self.logger.debug("CMD: ../gateways/ delete {} confirm {}".
                          format(gateway_name, confirm))

        self.logger.info("Deleting gateway, {}".format(gateway_name))

        confirm = self.ui_eval_param(confirm, 'bool', False)

        config = self.parent.parent.parent._get_config()
        if not config:
            self.logger.error("Unable to refresh local config over API - sync "
                              "aborted, restart rbd-target-api on {} to "
                              "sync".format(gateway_name))
            return

        target_iqn = self.parent.name

        gw_cnt = len(config['targets'][target_iqn]['portals'])
        if gw_cnt == 0:
            self.logger.error("Target is not mapped to any gateways.")
            return

        if gw_cnt == 1:
            if not confirm:
                self.logger.error("Deleting the last gateway will remove all "
                                  "objects on this target. Use confirm=true")
                return

        gw_api = '{}://{}:{}/api'.format(self.http_mode, "localhost",
                                         settings.config.api_port)
        gw_rqst = gw_api + '/gateway/{}/{}'.format(target_iqn, gateway_name)
        if confirm:
            gw_vars = {"force": 'true'}
        else:
            gw_vars = {"force": 'false'}

        api = APIRequest(gw_rqst, data=gw_vars)
        api.delete()

        msg = response_message(api.response, self.logger)
        if api.response.status_code != 200:
            if "unavailable:" + gateway_name in msg:
                self.logger.error("Could not contact {}. If the gateway is "
                                  "permanently down. Use confirm=true to "
                                  "force removal. WARNING: Forcing removal of "
                                  "a gateway that can still be reached by an "
                                  "initiator may result in data corruption.".
                                  format(gateway_name))
            else:
                self.logger.error("Failed : {}".format(msg))
            return

        self.logger.debug("{}".format(msg))
        self.logger.debug("Removing gw from UI")

        self.thread_lock.acquire()
        gw_object = self.get_child(gateway_name)
        self.remove_child(gw_object)
        self.thread_lock.release()

        config = self.parent.parent.parent._get_config()
        if not config:
            self.logger.error("Could not refresh display. Restart gwcli.")
        elif not config['targets'][target_iqn]['portals']:
            # no more gws so everything but the target is dropped.
            disks_object = self.parent.get_child("disks")
            disks_object.reset()

            hosts_grp_object = self.parent.get_child("host-groups")
            hosts_grp_object.reset()

            hosts_object = self.parent.get_child("hosts")
            hosts_object.reset()

    def ui_command_create(self, gateway_name, ip_addresses, nosync=False,
                          skipchecks='false'):
        """
        Define a gateway to the gateway group for this iscsi target. The
        first host added should be the gateway running the command

        gateway_name ... should resolve to the hostname of the gateway
        ip_addresses ... are the IPv4/IPv6 addresses of the interfaces the
                         iSCSI portals should use
        nosync ......... by default new gateways are sync'd with the
                         existing configuration by cli. By specifying nosync
                         the sync step is bypassed - so the new gateway
                         will need to have it's rbd-target-api daemon
                         restarted to apply the current configuration
                         (default = False)
        skipchecks ..... set this to true to force gateway validity checks
                         to be bypassed(default = false). This is a developer
                         option ONLY. Skipping these checks has the potential
                         to result in an unstable configuration.
        """

        ip_addresses = [normalize_ip_address(ip_address) for ip_address in ip_addresses.split(',')]
        self.logger.debug("CMD: ../gateways/ create {} {} "
                          "nosync={} skipchecks={}".format(gateway_name,
                                                           ip_addresses,
                                                           nosync,
                                                           skipchecks))

        local_gw = this_host()
        current_gateways = [tgt.name for tgt in self.children]

        if gateway_name != local_gw and len(current_gateways) == 0:
            # the first gateway defined must be the local machine. By doing
            # this the initial create uses localhost, and places it's portal IP
            # in the gateway ip list. Once the gateway ip list is defined, the
            # api server can resolve against the gateways - until the list is
            # defined only a request from localhost is acceptable to the api
            self.logger.error("The first gateway defined must be the local "
                              "machine")
            return

        if skipchecks not in ['true', 'false']:
            self.logger.error("skipchecks must be either true or false")
            return

        if local_gw in current_gateways:
            current_gateways.remove(local_gw)

        config = self.parent.parent.parent._get_config()
        if not config:
            self.logger.error("Unable to refresh local config"
                              " over API - sync aborted, restart rbd-target-api"
                              " on {} to sync".format(gateway_name))

        target_iqn = self.parent.name
        target_config = config['targets'][target_iqn]
        if nosync:
            sync_text = "sync skipped"
        else:
            sync_text = ("sync'ing {} disk(s) and "
                         "{} client(s)".format(len(target_config['disks']),
                                               len(target_config['clients'])))
        if skipchecks == 'true':
            self.logger.warning("OS version/package checks have been bypassed")

        self.logger.info("Adding gateway, {}".format(sync_text))

        gw_api = '{}://{}:{}/api'.format(self.http_mode,
                                         "localhost",
                                         settings.config.api_port)
        gw_rqst = gw_api + '/gateway/{}/{}'.format(target_iqn, gateway_name)
        gw_vars = {"nosync": nosync,
                   "skipchecks": skipchecks,
                   "ip_address": ','.join(ip_addresses)}

        api = APIRequest(gw_rqst, data=gw_vars)
        api.put()

        msg = response_message(api.response, self.logger)
        if api.response.status_code != 200:
            self.logger.error("Failed : {}".format(msg))
            return

        self.logger.debug("{}".format(msg))
        self.logger.debug("Adding gw to UI")

        # Target created OK, get the details back from the gateway and
        # add to the UI. We have to use the new gateway to ensure what
        # we get back is current (the other gateways will lag until they see
        # epoch xattr change on the config object)
        new_gw_endpoint = ('{}://{}:{}/'
                           'api'.format(self.http_mode,
                                        gateway_name,
                                        settings.config.api_port))

        api = APIRequest('{}/sysinfo/hostname'.format(new_gw_endpoint))
        api.get()
        gateway_hostname = api.response.json()['data']
        config = self.parent.parent.parent._get_config(endpoint=new_gw_endpoint)
        target_config = config['targets'][target_iqn]
        portal_config = target_config['portals'][gateway_hostname]
        Gateway(self, gateway_hostname, portal_config)

        self.logger.info('ok')

    def summary(self):

        up_count = len([gw.state for gw in self.children if gw.state == 'UP'])
        gw_count = len(self.children)

        return ("Up: {}/{}, Portals: {}".format(up_count,
                                                gw_count,
                                                gw_count),
                up_count == gw_count)

    @property
    def interactive(self):
        """determine whether the cli is running in interactive mode"""
        return self.parent.parent.parent.interactive


class Gateway(UINode):

    display_attributes = ["name",
                          "gateway_ip_list",
                          "portal_ip_addresses",
                          "inactive_portal_ips",
                          "tpgs",
                          "service_state"]

    TCP_PORT = 3260

    def __init__(self, parent, gateway_name, gateway_config):
        """
        Create the LIO element
        :param parent: parent object the gateway group object
        :param gateway_config: dict holding the fields that define the gateway
        :return:
        """

        UINode.__init__(self, gateway_name, parent)

        for k, v in gateway_config.items():
            self.__setattr__(k, v)

        self.state = "DOWN"
        self.service_state = {"iscsi": "DOWN",
                              "api": "DOWN"}

        self.refresh()

    def ui_command_refresh(self):
        """
        The refresh command will initiate a check against the gateway node,
        checking that the API is available, and that the iscsi port is
        listening
        """

        self.refresh()

    def refresh(self, mode="interactive"):

        if mode == 'interactive':
            self.logger.debug("- checking iSCSI/API ports on "
                              "{}".format(self.name))
        self._get_state()

    def _get_state(self):
        """
        Determine iSCSI and gateway API service state using the _ping api
        endpoint
        :return:
        """

        lookup = {200: {"status": "UP",
                        "iscsi": "UP", "api": "UP"},
                  401: {"status": "UNAUTHORIZED",
                        "iscsi": "UNKNOWN", "api": "UP"},
                  403: {"status": "UNAUTHORIZED",
                        "iscsi": "UNKNOWN", "api": "UP"},
                  500: {"status": "UNKNOWN",
                        "iscsi": "UNKNOWN", "api": "UNKNOWN"},
                  503: {"status": "PARTIAL",
                        "iscsi": "DOWN", "api": "UP"},
                  999: {"status": "UNKNOWN",
                        "iscsi": "UNKNOWN", "api": "UNKNOWN"},
                  }

        gw_api = '{}://{}:{}/api/_ping'.format(self.http_mode,
                                               self.name,
                                               settings.config.api_port)
        api = APIRequest(gw_api)
        try:
            api.get()
            rc = api.response.status_code
            if rc not in lookup:
                rc = 999
        except GatewayAPIError:
            rc = 999

        self.state = lookup[rc].get('status')
        self.service_state['iscsi'] = lookup[rc].get('iscsi')
        self.service_state['api'] = lookup[rc].get('api')

    def summary(self):

        state = self.state
        return "{} ({})".format(','.join(self.portal_ip_addresses),
                                state), (state == "UP")



from rtslib_fb import root
from rtslib_fb.utils import RTSLibError

__author__ = 'Paul Cuzner, Michael Christie'


class LIO(object):

    def __init__(self):
        self.lio_root = root.RTSRoot()
        self.error = False
        self.error_msg = ''
        self.changed = False

    def drop_lun_maps(self, config, update_config):

        disks_config = config.config['disks']
        backstore_object_names = [disk['backstore_object_name'] for _, disk in
                                  disks_config.items()]

        for stg_object in self.lio_root.storage_objects:

            if stg_object.name in backstore_object_names:

                # this is an rbd device that's in the config object,
                # so remove it
                try:
                    stg_object.delete()
                except RTSLibError as err:
                    self.error = True
                    self.error_msg = err
                else:
                    self.changed = True

                    if update_config:
                        # update the disk item to remove the wwn info
                        image_metadata = [disk for _, disk in config.config['disks'].items()
                                          if disk['backstore_object_name'] == stg_object.name][0]
                        image_metadata['wwn'] = ''
                        config.update_item("disks",
                                           stg_object.name,
                                           image_metadata)


class Gateway(LIO):

    def __init__(self, config_object):
        LIO.__init__(self)

        self.config = config_object

    def session_count(self):
        return len(list(self.lio_root.sessions))

    def drop_target(self, this_host):
        if this_host in self.config.config['gateways']:

            lio_root = root.RTSRoot()
            for tgt in lio_root.targets:
                if tgt.wwn in self.config.config['targets'] \
                        and this_host in self.config.config['targets'][tgt.wwn]['portals']:
                    tgt.delete()
                    self.changed = True



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import mock
import tempfile
import sys
import unittest
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding

# We need to mock ceph libs python bindings because there's no
# updated package in pypy
sys.modules['rados'] = mock.Mock()
sys.modules['rbd'] = mock.Mock()

from ceph_iscsi_config.client import CHAP  # noqa: E402
import ceph_iscsi_config.settings as settings  # noqa: E402
from base64 import b64encode  # noqa: E402


class ChapTest(unittest.TestCase):

    def setUp(self):
        settings.init()

    def test_chap_no_encryption(self):
        chap = CHAP("username", "password", False)
        self.assertEqual(chap.user, "username")
        self.assertEqual(chap.password, "password")
        self.assertEqual(chap.password_str, "password")

    def test_chap_encryption(self):
        private_key = rsa.generate_private_key(public_exponent=65537,
                                               key_size=2048,
                                               backend=default_backend())
        priv_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption())
        priv_key_file = tempfile.mkstemp()
        with open(priv_key_file[1], "wb") as kf:
            kf.write(priv_pem)

        pub_pem = private_key.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo)
        pub_key_file = tempfile.mkstemp()
        with open(pub_key_file[1], "wb") as kf:
            kf.write(pub_pem)

        settings.config.priv_key = priv_key_file[1]
        settings.config.pub_key = pub_key_file[1]
        settings.config.ceph_config_dir = ""

        chap = CHAP("username", "passwordverylonglong", False)

        encrypted_password = chap.encrypted_password(True)
        chap2 = CHAP(chap.user, encrypted_password, True)
        self.assertEqual(chap2.user, "username")
        self.assertEqual(chap2.password, "passwordverylonglong")
        self.assertEqual(chap2.password_str, encrypted_password)
        self.assertNotEqual(encrypted_password, "passwordverylonglong")

    def test_chap_upgrade(self):
        private_key = rsa.generate_private_key(public_exponent=65537,
                                               key_size=2048,
                                               backend=default_backend())
        priv_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption())
        priv_key_file = tempfile.mkstemp()
        with open(priv_key_file[1], "wb") as kf:
            kf.write(priv_pem)

        pub_pem = private_key.public_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo)
        pub_key_file = tempfile.mkstemp()
        with open(pub_key_file[1], "wb") as kf:
            kf.write(pub_pem)

        settings.config.priv_key = priv_key_file[1]
        settings.config.pub_key = pub_key_file[1]
        settings.config.ceph_config_dir = ""

        key = private_key.public_key()
        encrypted_pw = b64encode(key.encrypt("passwordverylonglong".encode('utf-8'),
                                 padding.OAEP(
                                     mgf=padding.MGF1(algorithm=hashes.SHA1()),
                                     algorithm=hashes.SHA1(),
                                     label=None))).decode('utf-8')

        chap2 = CHAP("username", encrypted_pw, True)
        self.assertEqual(chap2.user, "username")
        self.assertEqual(chap2.password, "passwordverylonglong")



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization, hashes
from cryptography.hazmat.primitives.asymmetric import padding
from base64 import b64encode, b64decode
import os

import rtslib_fb.root as lio_root

from rtslib_fb.target import NodeACL, Target, TPG
from rtslib_fb.fabric import ISCSIFabricModule
from rtslib_fb.utils import RTSLibError, RTSLibNotInCFS, normalize_wwn

import ceph_iscsi_config.settings as settings

from ceph_iscsi_config.gateway_setting import CLIENT_SETTINGS
from ceph_iscsi_config.common import Config
from ceph_iscsi_config.utils import encryption_available, CephiSCSIError, this_host
from ceph_iscsi_config.gateway_object import GWObject


class GWClient(GWObject):
    """
    This class holds a representation of a client connecting to LIO
    """
    SETTINGS = CLIENT_SETTINGS

    seed_metadata = {"auth": {"username": '',
                              "password": '',
                              "password_encryption_enabled": False,
                              "mutual_username": '',
                              "mutual_password": '',
                              "mutual_password_encryption_enabled": False},
                     "luns": {},
                     "group_name": ""
                     }

    def __init__(self, logger, client_iqn, image_list, username, password, mutual_username,
                 mutual_password, target_iqn):
        """
        Instantiate an instance of an LIO client
        :param client_iqn: (str) iscsi iqn string
        :param image_list: (list) list of rbd images (pool/image) to attach
                           to this client or list of tuples (disk, lunid)
        :param username: (str) chap username
        :param password: (str) chap password
        :param mutual_username: (str) chap mutual username
        :param mutual_password: (str) chap mutual password
        :param target_iqn: (str) target iqn string
        :return:
        """

        self.target_iqn = target_iqn
        self.lun_lookup = {}        # only used for hostgroup based definitions
        self.requested_images = []
        self.username = username
        self.password = password
        self.mutual_username = mutual_username
        self.mutual_password = mutual_password
        self.mutual = ''
        self.tpgauth = ''
        self.metadata = {}
        self.acl = None
        self.client_luns = {}
        self.tpg = None
        self.tpg_luns = {}
        self.lun_id_list = list(range(256))           # available LUN ids 0..255
        self.change_count = 0

        # enable commit to the config for changes by default
        self.commit_enabled = True

        self.logger = logger
        self.current_config = {}
        self.error = False
        self.error_msg = ''

        try:
            client_iqn, iqn_type = normalize_wwn(['iqn'], client_iqn)
        except RTSLibError as err:
            self.error = True
            self.error_msg = "Invalid iSCSI client name - {}".format(err)

        self.iqn = client_iqn

        # Validate the images list doesn't contain duplicate entries
        dup_images = set([rbd for rbd in image_list
                          if image_list.count(rbd) >= 2])
        if len(dup_images) > 0:
            self.error = True
            dup_string = ','.join(dup_images)
            self.error_msg = ("Client's image list contains duplicate rbd's"
                              ": {}".format(dup_string))

        try:
            super(GWClient, self).__init__('targets', target_iqn, logger,
                                           GWClient.SETTINGS)
        except CephiSCSIError as err:
            self.error = True
            self.error_msg = err

        # image_list is normally a list of strings (pool/image_name) but
        # group processing forces a specific lun id allocation to masked disks
        # in this scenario the image list is a tuple
        if image_list:

            if isinstance(image_list[0], tuple):
                # tuple format ('disk_name', {'lun_id': 0})...
                for disk_item in image_list:
                    disk_name = disk_item[0]
                    lun_id = disk_item[1].get('lun_id')
                    self.requested_images.append(disk_name)
                    self.lun_lookup[disk_name] = lun_id
            else:
                target_config = self.config.config['targets'][self.target_iqn]
                used_lun_ids = self._get_lun_ids(target_config['clients'])
                for disk_name in image_list:
                    disk_lun_id = target_config['disks'][disk_name]['lun_id']
                    if disk_lun_id not in used_lun_ids:
                        self.lun_lookup[disk_name] = disk_lun_id
                self.requested_images = image_list

    def _get_lun_ids(self, clients_config):
        lun_ids = []
        for client_config in clients_config.values():
            for lun_config in client_config['luns'].values():
                lun_ids.append(lun_config['lun_id'])
        return lun_ids

    def setup_luns(self, disks_config):
        """
        Add the requested LUNs to the node ACL definition. The image list
        defined for the client is compared to the current runtime settings,
        resulting in new images being added, or images removed.
        """

        # first drop the current lunid's used from the candidate list
        # this allows luns to be added/removed, and new id's to occupy free lun-id
        # slots rather than simply tag on the end. In a high churn environment,
        # adding new lun(s) at highest lun +1 could lead to exhausting the
        # 255 lun limit per target
        self.client_luns = self.get_images(self.acl)
        for image_name in self.client_luns:
            lun_id = self.client_luns[image_name]['lun_id']
            self.lun_id_list.remove(lun_id)
            self.logger.debug("(Client.setup_luns) {} has id of "
                              "{}".format(image_name, lun_id))

        self.tpg_luns = self.get_images(self.tpg)
        current_map = dict(self.client_luns)

        for image in self.requested_images:
            backstore_object_name = disks_config[image]['backstore_object_name']
            if backstore_object_name in self.client_luns:
                del current_map[backstore_object_name]
                continue
            else:
                rc = self._add_lun(image, self.tpg_luns[backstore_object_name])
                if rc != 0:
                    self.error = True
                    self.error_msg = ("{} is missing from the tpg - unable "
                                      "to map".format(image))
                    self.logger.debug("(Client.setup) tpg luns "
                                      "{}".format(self.tpg_luns))
                    self.logger.error("(Client.setup) missing image '{}' from "
                                      "the tpg".format(image))
                    return

        # 'current_map' should be empty, if not the remaining images need
        # to be removed from the client
        if current_map:
            for backstore_object_name in current_map:
                self._del_lun_map(backstore_object_name, disks_config)
                if self.error:
                    self.logger.error("(Client.setup) unable to delete {} from"
                                      " {}".format(self.iqn,
                                                   backstore_object_name))
                    return

    def update_acl_controls(self):
        self.logger.debug("(update_acl_controls) controls: {}".format(self.controls))
        self.acl.set_attribute('dataout_timeout', str(self.dataout_timeout))

        # Try to detect network problems so we can kill connections
        # and cleanup before the initiator has begun recovery and
        # failed over.

        # LIO default 30
        self.acl.set_attribute('nopin_response_timeout',
                               str(self.nopin_response_timeout))
        # LIO default 15
        self.acl.set_attribute('nopin_timeout', str(self.nopin_timeout))

        # LIO default 64
        self.acl.tcq_depth = self.cmdsn_depth

    def define_client(self):
        """
        Establish the links for this object to the corresponding ACL and TPG
        objects from LIO
        :return:
        """

        iscsi_fabric = ISCSIFabricModule()
        target = Target(iscsi_fabric, self.target_iqn, 'lookup')

        # NB. this will check all tpg's for a matching iqn
        for tpg in target.tpgs:
            if tpg.enable:
                for client in tpg.node_acls:
                    if client.node_wwn == self.iqn:
                        self.acl = client
                        self.tpg = client.parent_tpg
                        try:
                            self.update_acl_controls()
                        except RTSLibError as err:
                            self.logger.error("(Client.define_client) FAILED to update "
                                              "{}".format(self.iqn))
                            self.error = True
                            self.error_msg = err
                        self.logger.debug("(Client.define_client) - {} already "
                                          "defined".format(self.iqn))
                        return

        # at this point the client does not exist, so create it
        # The configuration only has one active tpg, so pick that one for any
        # acl definitions
        for tpg in target.tpgs:
            if tpg.enable:
                self.tpg = tpg

        try:
            self.acl = NodeACL(self.tpg, self.iqn)
            self.update_acl_controls()
        except RTSLibError as err:
            self.logger.error("(Client.define_client) FAILED to define "
                              "{}".format(self.iqn))
            self.logger.debug("(Client.define_client) failure msg "
                              "{}".format(err))
            self.error = True
            self.error_msg = err
        else:
            self.logger.info("(Client.define_client) {} added "
                             "successfully".format(self.iqn))
            self.change_count += 1

    @staticmethod
    def get_client_info(target_iqn, client_iqn):
        result = {
            "alias": '',
            "state": '',
            "ip_address": []
        }
        iscsi_fabric = ISCSIFabricModule()
        try:
            target = Target(iscsi_fabric, target_iqn, 'lookup')
        except RTSLibNotInCFS:
            return result
        for tpg in target.tpgs:
            if tpg.enable:
                for client in tpg.node_acls:
                    if client.node_wwn != client_iqn:
                        continue
                    session = client.session
                    if session is None:
                        break
                    result['alias'] = session.get('alias')
                    state = session.get('state').upper()
                    result['state'] = state
                    ips = set()
                    if state == 'LOGGED_IN':
                        for conn in session.get('connections'):
                            ips.add(conn.get('address'))
                        result['ip_address'] = list(ips)
                    break
        return result

    @staticmethod
    def define_clients(logger, config, target_iqn):
        """
        define the clients (nodeACLs) to the gateway definition
        :param logger: logger object to print to
        :param config: configuration dict from the rados pool
        :raises CephiSCSIError.
        """

        # Client configurations (NodeACL's)
        target_config = config.config['targets'][target_iqn]
        for client_iqn in target_config['clients']:
            client_metadata = target_config['clients'][client_iqn]
            client_chap = CHAP(client_metadata['auth']['username'],
                               client_metadata['auth']['password'],
                               client_metadata['auth']['password_encryption_enabled'])
            client_chap_mutual = CHAP(client_metadata['auth']['mutual_username'],
                                      client_metadata['auth']['mutual_password'],
                                      client_metadata['auth'][
                                          'mutual_password_encryption_enabled'])

            image_list = list(client_metadata['luns'].keys())

            if client_chap.error:
                raise CephiSCSIError("Unable to decode password for {}. "
                                     "CHAP error: {}".format(client_iqn,
                                                             client_chap.error_msg))
            if client_chap_mutual.error:
                raise CephiSCSIError("Unable to decode password for {}. "
                                     "CHAP_MUTUAL error: {}".format(client_iqn,
                                                                    client_chap.error_msg))

            client = GWClient(logger,
                              client_iqn,
                              image_list,
                              client_chap.user,
                              client_chap.password,
                              client_chap_mutual.user,
                              client_chap_mutual.password,
                              target_iqn)

            client.manage('present')  # ensure the client exists

    @staticmethod
    def try_disable_auth(tpg):
        """
        Disable authentication (enable ACL mode) if this is the last CHAP user.

        LIO doesn't allow us to mix and match ACLs and auth under a tpg. We
        only allow ACL mode if there are not CHAP users.
        """

        for client in tpg.node_acls:
            if client.chap_userid or client.chap_password:
                return

        if tpg.chap_userid or tpg.chap_password:
            return

        tpg.set_attribute('authentication', '0')

    def configure_auth(self, username, password, mutual_username, mutual_password, target_config):
        """
        Attempt to configure authentication for the client
        :return:
        """

        auth_enabled = (username and password)

        self.logger.debug("configuring auth username={}, password={}, mutual_username={}, "
                          "mutual_password={}".format(username, password, mutual_username,
                                                      mutual_password))
        acl_chap_userid = self.acl.chap_userid
        acl_chap_password = self.acl.chap_password
        acl_chap_mutual_userid = self.acl.chap_mutual_userid
        acl_chap_mutual_password = self.acl.chap_mutual_password

        try:
            self.logger.debug("Updating the ACL")
            if username != acl_chap_userid or \
                    password != acl_chap_password:
                self.acl.chap_userid = username
                self.acl.chap_password = password

                new_chap = CHAP(username, password, False)
                self.logger.debug("chap object set to: {},{},{}".format(
                    new_chap.user, new_chap.password, new_chap.password_str))

                if new_chap.error:
                    self.error = True
                    self.error_msg = new_chap.error_msg
                    return

            if mutual_username != acl_chap_mutual_userid or \
                    mutual_password != acl_chap_mutual_password:
                self.acl.chap_mutual_userid = mutual_username
                self.acl.chap_mutual_password = mutual_password

                new_chap_mutual = CHAP(mutual_username, mutual_password, False)
                self.logger.debug("chap mutual object set to: {},{},{}".format(
                    new_chap_mutual.user, new_chap_mutual.password,
                    new_chap_mutual.password_str))

                if new_chap_mutual.error:
                    self.error = True
                    self.error_msg = new_chap_mutual.error_msg
                    return

            if auth_enabled:
                self.tpg.set_attribute('authentication', '1')
            else:
                GWClient.try_disable_auth(self.tpg)

            self.logger.debug("Updating config object meta data")
            encryption_enabled = encryption_available()
            if username != acl_chap_userid:
                self.metadata['auth']['username'] = new_chap.user
            if password != acl_chap_password:
                self.metadata['auth']['password'] = new_chap.encrypted_password(encryption_enabled)
                self.metadata['auth']['password_encryption_enabled'] = encryption_enabled
            if mutual_username != acl_chap_mutual_userid:
                self.metadata['auth']['mutual_username'] = new_chap_mutual.user
            if mutual_password != acl_chap_mutual_password:
                self.metadata['auth']['mutual_password'] = \
                    new_chap_mutual.encrypted_password(encryption_enabled)
                self.metadata['auth']['mutual_password_encryption_enabled'] = encryption_enabled

        except RTSLibError as err:
            self.error = True
            self.error_msg = ("Unable to configure authentication "
                              "for {} - ".format(self.iqn,
                                                 err))
            self.logger.error("(Client.configure_auth) failed to set "
                              "credentials for {}".format(self.iqn))
        else:
            self.change_count += 1

        self._update_acl(target_config)

    def _update_acl(self, target_config):
        if self.tpg.node_acls:
            self.tpg.set_attribute('generate_node_acls', 0)
            self.tpg.set_attribute('demo_mode_write_protect', 1)
            if not target_config['acl_enabled']:
                target_config['acl_enabled'] = True
                self.change_count += 1

    def _add_lun(self, image, lun):
        """
        Add a given image to the client ACL
        :param image: rbd image name of the form pool/image (str)
        :param lun: rtslib lun object
        :return:
        """

        rc = 0
        # get the tpg lun to map this client to
        tpg_lun = lun['tpg_lun']

        # lunid allocated from the current config object setting, or if this is
        # a new device from the target disk lun id or next free lun id 'position'
        # if target disk lun id is already in use
        if image in self.metadata['luns'].keys():
            lun_id = self.metadata['luns'][image]['lun_id']
        else:
            if image in self.lun_lookup:
                lun_id = self.lun_lookup[image]
            else:
                lun_id = self.lun_id_list[0]  # pick lowest available lun ID

        self.logger.debug("(Client._add_lun) Adding {} to {} at "
                          "id {}".format(image, self.iqn, lun_id))

        try:
            m_lun = self.acl.mapped_lun(lun_id, tpg_lun=tpg_lun)
        except RTSLibError as err:
            self.logger.error("Client.add_lun RTSLibError for lun id {} -"
                              " {}".format(lun_id, err))
            rc = 12
        else:

            self.client_luns[image] = {"lun_id": lun_id,
                                       "mapped_lun": m_lun,
                                       "tpg_lun": tpg_lun}

            self.metadata['luns'][image] = {"lun_id": lun_id}
            self.lun_id_list.remove(lun_id)
            self.logger.info("(Client.add_lun) added image '{}' to "
                             "{}".format(image, self.iqn))
            self.change_count += 1

        return rc

    def _del_lun_map(self, backstore_object_name, disks_config):
        """
        Delete a lun from the client's ACL
        :param backstore_object_name: rbd image name to remove
        :return:
        """

        lun = self.client_luns[backstore_object_name]['mapped_lun']
        try:
            lun.delete()
        except RTSLibError as err:
            self.error = True
            self.error_msg = err
        else:
            self.change_count += 1

            disk_id = [disk_id for disk_id, disk in disks_config.items()
                       if disk['backstore_object_name'] == backstore_object_name][0]
            # the lun entry could have been deleted by another host, so before
            # we try and delete - make sure it's in our local copy of the
            # metadata!
            if disk_id in self.metadata['luns']:
                del self.metadata['luns'][disk_id]

    def delete(self):
        """
        Delete the client definition from LIO
        :return:
        """

        try:
            self.acl.delete()
            GWClient.try_disable_auth(self.tpg)
            self.change_count += 1
            self.logger.info("(Client.delete) deleted NodeACL for "
                             "{}".format(self.iqn))
        except RTSLibError as err:
            self.error = True
            self.error_msg = "RTS NodeACL delete failure"
            self.logger.error("(Client.delete) failed to delete client {} "
                              "- error: {}".format(self.iqn,
                                                   err))

    def exists(self):
        """
        This function determines whether this instances iqn is already defined
        to LIO
        :return: Boolean
        """

        r = lio_root.RTSRoot()
        client_list = [client.node_wwn for client in r.node_acls]
        return self.iqn in client_list

    def seed_config(self, config):
        """
        function to seed the config object with a new client definition
        """
        target_config = config.config["targets"][self.target_iqn]
        target_config['clients'][self.iqn] = GWClient.seed_metadata
        config.update_item("targets", self.target_iqn, target_config)

        # persist the config update, and leave the connection to the ceph
        # object open since adding just the iqn is only the start of the
        # definition
        config.commit("retain")

    def manage(self, rqst_type, committer=None):
        """
        Manage the allocation or removal of this client
        :param rqst_type is either 'present' (try and create the nodeACL), or
        'absent' - delete the nodeACL
        :param committer is the host responsible for any commits to the
        configuration - this is not needed for Ansible management, but is used
        by the CLI->API->GWClient interaction
        """
        # Build a local object representing the rados configuration object
        config_object = Config(self.logger)
        if config_object.error:
            self.error = True
            self.error_msg = config_object.error_msg
            return

        # use current config to hold a copy of the current rados config
        # object (dict)
        self.current_config = config_object.config
        target_config = self.current_config['targets'][self.target_iqn]
        update_host = committer

        self.logger.debug("GWClient.manage) update host to handle any config "
                          "update is {}".format(update_host))

        if rqst_type == "present":

            ###################################################################
            # Ensure the client exists in LIO                                 #
            ###################################################################

            # first look at the request to see if it matches the settings
            # already in the config object - if so this is just a rerun, or a
            # reboot so config object updates are not needed when we change
            # the LIO environment
            if self.iqn in target_config['clients'].keys():
                self.metadata = target_config['clients'][self.iqn]
                config_image_list = sorted(self.metadata['luns'].keys())

                #
                # Does the request match the current config?

                auth_config = self.metadata['auth']
                config_chap = CHAP(auth_config['username'],
                                   auth_config['password'],
                                   auth_config['password_encryption_enabled'])
                if config_chap.error:
                    self.error = True
                    self.error_msg = config_chap.error_msg
                    return
                # extract the chap_mutual_str from the config object entry
                config_chap_mutual = CHAP(auth_config['mutual_username'],
                                          auth_config['mutual_password'],
                                          auth_config['mutual_password_encryption_enabled'])
                if config_chap_mutual.error:
                    self.error = True
                    self.error_msg = config_chap_mutual.error_msg
                    return

                if self.username == config_chap.user and \
                   self.password == config_chap.password and \
                   self.mutual_username == config_chap_mutual.user and \
                   self.mutual_password == config_chap_mutual.password and \
                   config_image_list == sorted(self.requested_images):
                    self.commit_enabled = False
            else:
                # requested iqn is not in the config object
                self.seed_config(config_object)
                self.metadata = GWClient.seed_metadata

            self.logger.debug("(manage) config updates to be applied from "
                              "this host: {}".format(self.commit_enabled))

            client_exists = self.exists()
            self.define_client()
            if self.error:
                # unable to define the client!
                return

            if client_exists and self.metadata["group_name"]:
                # bypass setup_luns for existing clients that have an
                # associated host group
                pass
            else:
                # either the client didn't exist (new or boot time), or the
                # group_name is not defined so run setup_luns for this client
                disks_config = self.current_config['disks']
                bad_images = self.validate_images(disks_config)
                if not bad_images:

                    self.setup_luns(disks_config)
                    if self.error:
                        return
                else:
                    # request for images to map to this client that haven't
                    # been added to LIO yet!
                    self.error = True
                    self.error_msg = ("Non-existent images {} requested "
                                      "for {}".format(bad_images, self.iqn))
                    return

            if not self.username and not self.password and \
               not self.mutual_username and not self.mutual_password:
                self.logger.warning("(main) client '{}' configured without"
                                    " security".format(self.iqn))

            self.configure_auth(self.username, self.password, self.mutual_username,
                                self.mutual_password, target_config)

            if self.error:
                return

            # check the client object's change count, and update the config
            # object if this is the updating host
            if self.change_count > 0:

                if self.commit_enabled:

                    if update_host == this_host():
                        # update the config object with this clients settings
                        self.logger.debug("Updating config object metadata "
                                          "for '{}'".format(self.iqn))
                        target_config['clients'][self.iqn] = self.metadata
                        config_object.update_item("targets",
                                                  self.target_iqn,
                                                  target_config)

                        # persist the config update
                        config_object.commit()

        elif rqst_type == 'reconfigure':
            self.define_client()

        else:
            ###################################################################
            # Remove the requested client from the config object and LIO      #
            ###################################################################
            if self.exists():
                self.define_client()   # grab the client and parent tpg objects
                self.delete()          # deletes from the local LIO instance
                if self.error:
                    return
                else:
                    # remove this client from the config

                    if update_host == this_host():
                        self.logger.debug("Removing {} from the config "
                                          "object".format(self.iqn))
                        target_config['clients'].pop(self.iqn)
                        config_object.update_item("targets", self.target_iqn, target_config)
                        config_object.commit()

            else:
                # desired state is absent, but the client does not exist
                # in LIO - Nothing to do!
                self.logger.info("(main) client {} removal request, but it's"
                                 "not in LIO...skipping".format(self.iqn))

    def validate_images(self, disks_config):
        """
        Confirm that the images listed are actually allocated to the tpg and
        can therefore be used by a client
        :return: a list of images that are NOT in the tpg ... should be empty!
        """
        bad_images = []
        tpg_lun_list = self.get_images(self.tpg).keys()
        self.logger.debug("tpg images: {}".format(tpg_lun_list))
        self.logger.debug("request images: {}".format(self.requested_images))
        backstore_object_names = [disk['backstore_object_name'] for disk_id, disk
                                  in disks_config.items()
                                  if disk_id in self.requested_images]
        self.logger.debug("backstore object names: {}".format(backstore_object_names))
        for backstore_object_name in backstore_object_names:
            if backstore_object_name not in tpg_lun_list:
                bad_images.append(backstore_object_name)

        return bad_images

    @staticmethod
    def get_update_host(config):
        """
        decide which gateway host should be responsible for any config object
        updates
        :param config: configuration dict from the rados pool
        :return: a suitable gateway host that is online
        """

        ptr = 0
        potential_hosts = [host_name for host_name in config["gateways"].keys()
                           if isinstance(config["gateways"][host_name], dict)]

        # Assume the 1st element from the list is OK for now
        # TODO check the potential hosts are online/available

        return potential_hosts[ptr]

    def get_images(self, rts_object):
        """
        Funtion to return a dict of luns mapped to either a node ACL or the
        TPG, based on the passed object type
        :param rts_object: rtslib object - either NodeACL or TPG
        :return: dict indexed by image name of LUN object attributes
        """

        luns_mapped = {}

        if isinstance(rts_object, NodeACL):
            # return a dict of images assigned to this client
            for m_lun in rts_object.mapped_luns:

                key = m_lun.tpg_lun.storage_object.name
                luns_mapped[key] = {"lun_id": m_lun.mapped_lun,
                                    "mapped_lun": m_lun,
                                    "tpg_lun": m_lun.tpg_lun}

        elif isinstance(rts_object, TPG):
            # return a dict of *all* images available to this tpg
            for m_lun in rts_object.luns:

                key = m_lun.storage_object.name
                luns_mapped[key] = {"lun_id": m_lun.lun,
                                    "mapped_lun": None,
                                    "tpg_lun": m_lun}
        return luns_mapped


class CHAP(object):

    def __init__(self, user, password_str, encryption_enabled):

        self.error = False
        self.error_msg = ''

        self.user = user
        self.password_str = password_str
        if len(self.password_str) > 0 and encryption_enabled:
            self.password = self._decrypt()
        else:
            self.password = self.password_str

    def encrypted_password(self, encryption_enabled):
        if encryption_enabled and len(self.password_str) > 0:
            return self._encrypt()
        return self.password

    def _decrypt(self):

        key_path = os.path.join(settings.config.ceph_config_dir,
                                settings.config.priv_key)
        try:
            with open(key_path, 'rb') as keyf:
                key = serialization.load_pem_private_key(keyf.read(), None,
                                                         default_backend())
            try:
                plain_pw = key.decrypt(b64decode(self.password_str),
                                       padding.OAEP(
                                           mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                           algorithm=hashes.SHA256(),
                                           label=None)).decode('utf-8')
            except ValueError:
                # decrypting a password that was encrypted with python-crypto?
                plain_pw = key.decrypt(b64decode(self.password_str),
                                       padding.OAEP(
                                           mgf=padding.MGF1(algorithm=hashes.SHA1()),
                                           algorithm=hashes.SHA1(),
                                           label=None)).decode('utf-8')
        except Exception as ex:
            print(ex)
            self.error = True
            self.error_msg = 'Problems decoding the encrypted password'
            return None
        else:
            return plain_pw

    def _encrypt(self):
        key_path = os.path.join(settings.config.ceph_config_dir,
                                settings.config.pub_key)
        try:
            with open(key_path, 'rb') as keyf:
                key = serialization.load_pem_public_key(keyf.read(),
                                                        default_backend())

            encrypted_pw = b64encode(key.encrypt(self.password_str.encode('utf-8'),
                                     padding.OAEP(
                                         mgf=padding.MGF1(algorithm=hashes.SHA256()),
                                         algorithm=hashes.SHA256(),
                                         label=None))).decode('utf-8')
        except Exception:
            self.error = True
            self.error_msg = 'Encoding password failed'
            return None
        else:
            return encrypted_pw
